/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "apocalyx/glapp.h"
#include "apocalyx/glparticle.h"
#include "apocalyx/glmesh.h"
#include "apocalyx/glterrain.h"
#include "apocalyx/glsprite.h"
#include "apocalyx/dataread.h"
#include "apocalyx/slinterpreter.h"
#include "apocalyx/datasets.cpp"

//
// BasicScene
//
class BasicScene: public GLScene {
public:
  enum {
    SPLASH_SCENE = 0,
    INTRO_SCENE = 1,
    GAME_SCENE = 2,
  };
public:
  virtual char* getHelpLine(int line) = 0;
};

//
// SplashScene
//
class SplashScene: public BasicScene {
private:
  DRImage* splashImage;
public:
  SplashScene();
  bool initialize(GLWin& glWin);
  void finalize(GLWin& glWin);
  int update(GLWin& glWin);
  char* getHelpLine(int line);
};

SplashScene::SplashScene(): splashImage(NULL) {
}

bool SplashScene::initialize(GLWin& glWin) {
  glWin.setHelpMode(GLWin::USER_HELP);
  glWin.setShowFramerate(false);
  glWin.getCamera().setOrtho();
  try {
    DRZipFile zip("ZekeOnYourSix!.dat");
    splashImage = zip.getImage("loading.jpg");
  } catch(Exception& e) {
    return false;
  }
  return true;
}

void SplashScene::finalize(GLWin& glWin) {
  if(splashImage) {
    delete splashImage;
    splashImage = NULL;
  }
  glWin.setHelpMode(GLWin::REDUCED_HELP);
  glWin.setShowFramerate(true);
}

int SplashScene::update(GLWin& glWin) {
  const static float startTime = GLWin::getElapsedTime();
  const float waitTime = 10.0f;
  glClear(GL_COLOR_BUFFER_BIT);
  if(splashImage) {
    int width = glWin.getWidth();
    int height = glWin.getHeight();
    int imgWidth = splashImage->getWidth();
    int imgHeight = splashImage->getHeight();
    glRasterPos2f((float)imgWidth/width,-(float)imgHeight/height);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glDrawPixels(
      imgWidth,imgHeight,GL_RGB,GL_UNSIGNED_BYTE,splashImage->getBuffer()
    );
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
  }
  if(glWin.isKeyPressed('1') || GLWin::getElapsedTime()-startTime > waitTime) {
    glWin.keyUp('1');
    return INTRO_SCENE;
  } else if(glWin.isKeyPressed('2')) {
    glWin.keyUp('2');
    return GAME_SCENE;
  }
  return SPLASH_SCENE;
}

char* SplashScene::getHelpLine(int line) {
  switch(line) {
    case 1: return "[1] Start Intro";
    case 0: return "[2] Start Fly-Through";
    default: return NULL;
  }
}

//
// IntroScene
//
class IntroScene: public BasicScene {
protected:
  enum Sequence {
    SCROLLING_TEXTS = 0,
    SUN_RISE,
    SHOW_ISLES,
    APPROACH_ISLES,
    APPROACH_ISLES_2,
    WATCH_TREES,
    WATCH_CARRIER,
    FLY_THROUGH,
  };
private:
  FMMusic *soundTrack;
  FMSound *waterSound, *windSound;
  FMSample *waterSample, *windSample;
  FM3DSample *alarmSample, *fireSample;
  FM3DSample *motorSample;
  GLBackground *starBackground, *skyBackground;
  GLParagraph *scrollingParagraph, *slidingParagraph;
  GLFlatTerrain* sea;
  GLSunLight* sun;
  GLHeightField *isle[3];
  GLTrees *trees[3];
  GLObjects *carrier, *zero[3], *wild[3];
  GLEmitter *fireEmitter[2], *smokeEmitter[2];
  float sequenceTime;
  Sequence sequence;
protected:
  void generateIsle(
    GLWin& glWin, DRZipFile& zip, char* txtName, GLTexture* detail,
    int detailRepeat, char* hghtName, float waterLev, float width, float depth,
    float height, float x, float y, int treesCount, GLMaterial* treesMaterial,
    float minTreesH, GLHeightField** isle, GLTrees** trees
  );
  GLMaterial* generateTreesMaterial(DRZipFile& zip, char* fileName);
  GLObjects* cloneMeshes(GLObjects& objs);
public:
  IntroScene();
  bool initialize(GLWin& glWin);
  void finalize(GLWin& glWin);
  int update(GLWin& glWin);
  char* getHelpLine(int line);
};

IntroScene::IntroScene():
  soundTrack(NULL), waterSound(NULL), waterSample(NULL),
  alarmSample(NULL), fireSample(NULL),
  motorSample(NULL),
  windSound(NULL), windSample(NULL),
  starBackground(NULL), skyBackground(NULL),
  scrollingParagraph(NULL), slidingParagraph(NULL),
  sun(NULL), sea(NULL),
  carrier(NULL),
  sequenceTime(0),
  sequence(SCROLLING_TEXTS)
{
  for(int ct = 0; ct < 3; ct++) {
    isle[ct] = NULL;
    trees[ct] = NULL;
    zero[ct] = NULL;
    wild[ct] = NULL;
  }
  for(int ct = 0; ct < 2; ct++) {
    fireEmitter[ct] = NULL;
    smokeEmitter[ct] = NULL;
  }
}

bool IntroScene::initialize(GLWin& glWin) {
  glWin.setHelpMode(GLWin::HIDDEN_HELP);
  glWin.setPerspective(60,0.5f,3000);
  glWin.getCamera().setPosition(0,10,0);
  glWin.getWorld().setAmbient(1,1,1);
  try {
    DRZipFile zip("ZekeOnYourSix!.dat");
    /// background
    {
      const int texturesCount = 5;
      DRImage* skyboxImages[texturesCount] = {
        zip.getImage("SkyboxTop.jpg"), zip.getImage("SkyboxLeft.jpg"),
        zip.getImage("SkyboxFront.jpg"), zip.getImage("SkyboxRight.jpg"),
        zip.getImage("SkyboxBack.jpg"),
      };
      GLTexture* skyboxTextures[texturesCount];
      for(int ct = 0; ct < texturesCount; ct++)
        skyboxTextures[ct] = new GLTexture(*skyboxImages[ct]);
      for(int ct = 0; ct < texturesCount; ct++) delete skyboxImages[ct];
      skyBackground = new GLMirroredSkybox(skyboxTextures);
      skyBackground->setColor(0,0,0);
      skyBackground->rotateStanding(3.14159);

      DRImage* backImage = zip.getImage("stars.jpg");
      GLTexture* backTexture = new GLTexture(*backImage);
      delete backImage;
      starBackground =
        new GLStarfield(50,6000,backTexture,20);
      glWin.getWorld().setBackground(starBackground);
    }
    /// text
    {
      GLOutlineFont* verdana = new GLOutlineFont(
        glWin.getHDC(),"Verdana",24,0,0.15f,FW_BOLD
      );
      GLOutlineFont* courier = new GLOutlineFont(
        glWin.getHDC(),"Courier New",24,0,0.2f,FW_BOLD
      );
      // sliding paragraph
      {
        char* texts[] = {
          "A long time ago,              ",
          "in a galaxy NOT SO far away...",
          NULL,
        };
        GLMaterial* textMaterial = new GLMaterial();
        textMaterial->setAmbient(0.25f,0.25f,0.5f);
        textMaterial->setDiffuse(0.25f,0.25f,1);
        slidingParagraph = new GLParagraph();
        int ct = 0;
        for(; texts[ct] != NULL; ct++) {
          GLText* text = new GLText(texts[ct],courier,textMaterial);
          text->rotateStanding(3.1415f);
          text->move(4,-ct,0);
          slidingParagraph->addText(text);
        }
        slidingParagraph->setVisible(true);
        slidingParagraph->move(0,10,16);
        glWin.getWorld().addObject(slidingParagraph);
      }
      // scrolling paragraph
      {
        char* texts[] = {
          "Four Elements Demo",
          "Episode III",
          "\"Zeke On Your Six!\"",
          " ",
          "The last effort to strike the",
          "Republican army has descreased",
          "the power of the Empire.",
          " ",
          "The final battle is near and the",
          "imperial fighters are searching for",
          "the republican carriers near three",
          "rocks in the Delta quadrant.",
          " ",
          "The pilots of both sides know that",
          "this skirmish is decisive and",
          "they'll fight with great valour.",
          NULL,
        };
        GLMaterial* textMaterial = new GLMaterial();
        textMaterial->setAmbient(0.25f,0.25f,0.25f);
        textMaterial->setDiffuse(1,1,0);
        scrollingParagraph = new GLParagraph();
        int ct = 0;
        for(; texts[ct] != NULL; ct++) {
          GLText* text = new GLText(texts[ct],verdana,textMaterial);
          text->rotateStanding(3.1415f);
          text->move(0,-ct,0);
          scrollingParagraph->addText(text);
        }
        scrollingParagraph->setVisible(true);
        scrollingParagraph->pitch(1.047);
        scrollingParagraph->move(0,1,14);
        glWin.getWorld().addObject(scrollingParagraph);
      }
      /// sun
      {
        DRImage* sunImage = zip.getImage("light.jpg");
        GLTexture* sunTexture = new GLTexture(*sunImage);
        delete sunImage;
        DRImage* flareImage = zip.getImage("lensflares.png");
        GLTexture* flareTexture = new GLTexture(*flareImage);
        delete flareImage;
        M3Vector direc(0,0.413793f,-0.910345);
        sun = new GLSunLight(
          sunTexture,0.25f,direc,flareTexture,7,0.2f
        );
        sun->setVisible(true);
        sun->setColor(1,1,1);
        glWin.getWorld().setSun(sun);
      }
      /// sea
      {
        const int imagesCount = 16;
        DRImage* waterImages[imagesCount] = {
          zip.getImage("water00.jpg"),
          zip.getImage("water01.jpg"),
          zip.getImage("water02.jpg"),
          zip.getImage("water03.jpg"),
          zip.getImage("water04.jpg"),
          zip.getImage("water05.jpg"),
          zip.getImage("water06.jpg"),
          zip.getImage("water07.jpg"),
          zip.getImage("water08.jpg"),
          zip.getImage("water09.jpg"),
          zip.getImage("water10.jpg"),
          zip.getImage("water11.jpg"),
          zip.getImage("water12.jpg"),
          zip.getImage("water13.jpg"),
          zip.getImage("water14.jpg"),
          zip.getImage("water15.jpg"),
        };
        GLAnimatedTexture* waterTexture =
          new GLAnimatedTexture(imagesCount,waterImages,1.5f,true);
        for(int ct = 0; ct < imagesCount; ct++) delete waterImages[ct];
        GLMaterial* terrainMaterial = new GLMaterial();
        terrainMaterial->setAmbient(0.7f,0.7f,0.7f);
        terrainMaterial->setDiffuse(0,0.5f,0.9f);
        terrainMaterial->setDiffuseTexture(waterTexture);
        sea = new GLFlatTerrain(terrainMaterial,3000,200);
        sea->setVisible(false);
        sea->setReflective(true);
        glWin.getWorld().setFlatTerrain(sea);
      }
    }
    /// 3DS
    {
      carrier = zip.getMeshes("carrier.3ds");
      carrier->move(600,-2,300);
      carrier->pitch(-3.1415/2);
      carrier->rotateStanding(-3.1415/2);
      carrier->setVisible(false);
      glWin.getWorld().addObject(carrier);
      zero[0] = zip.getMeshes("zero.3ds");
      zero[0]->move(-300,15,300);
      zero[0]->pitch(-3.1415/2);
      zero[0]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(zero[0]);
      zero[1] = cloneMeshes(*zero[0]);
      zero[1]->move(-270,55,320);
      zero[1]->pitch(-3.1415/2);
      zero[1]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(zero[1]);
      zero[2] = cloneMeshes(*zero[0]);
      zero[2]->move(-280,50,310);
      zero[2]->pitch(-3.1415/2);
      zero[2]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(zero[2]);
      wild[0] = zip.getMeshes("wildcat.3ds");
      wild[0]->move(-500,35,310);
      wild[0]->pitch(-3.1415/2);
      wild[0]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(wild[0]);
      wild[1] = cloneMeshes(*wild[0]);
      wild[1]->move(-510,25,320);
      wild[1]->pitch(-3.1415/2);
      wild[1]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(wild[1]);
      wild[2] = cloneMeshes(*wild[0]);
      wild[2]->move(-525,65,330);
      wild[2]->pitch(-3.1415/2);
      wild[2]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(wild[2]);
    }
    /// emitters
    {
      DRImage* fireImage = zip.getImage("fire.png");
      fireImage->convertGray2111A();
      GLTexture* fireTexture = new GLTexture(*fireImage);
      DRImage* smokeImage = zip.getImage("smoke.png");
      smokeImage->convertGray2RGBA();
      GLTexture* smokeTexture = new GLTexture(*smokeImage);
      for(int ct = 0; ct < 2; ct++) {
        fireEmitter[ct] = new GLEmitter(10,1,15,true);
        fireEmitter[ct]->setParticleTexture(
          fireTexture,fireImage->getHasAlpha()
        );
        fireEmitter[ct]->setParticleVelocity(M3Vector(-3,2,0),0.5f);
        fireEmitter[ct]->setColor(GLColor(1,1,0,1),GLColor(1,0,0,0));
        fireEmitter[ct]->setSize(2,.5);
        fireEmitter[ct]->setGravity(M3Vector(),M3Vector());
        fireEmitter[ct]->setVisible(false);
        glWin.getWorld().addObject(fireEmitter[ct]);
        smokeEmitter[ct] = new GLEmitter(20,4,30,true);
        smokeEmitter[ct]->setParticleTexture(
          smokeTexture,smokeImage->getHasAlpha()
        );
        smokeEmitter[ct]->setParticleVelocity(M3Vector(-4,1,0),0.5f);
        smokeEmitter[ct]->setColor(GLColor(0,0,0,1),GLColor(0,0,0,0));
        smokeEmitter[ct]->setSize(.25,2);
        smokeEmitter[ct]->setGravity(M3Vector(),M3Vector(0,-4,0));
        smokeEmitter[ct]->setVisible(false);
        glWin.getWorld().addObject(smokeEmitter[ct]);
      }
      delete fireImage;
      delete smokeImage;
      fireEmitter[0]->move(-514,25,320);
      smokeEmitter[0]->move(-514,25,320);
      fireEmitter[1]->move(-304,15,300);
      smokeEmitter[1]->move(-304,15,300);
    }
    /// isles
    {
      GLMaterial* treesMaterial1 = generateTreesMaterial(zip,"trees1.png");
      GLMaterial* treesMaterial2 = generateTreesMaterial(zip,"trees2.png");
      GLMaterial* treesMaterial3 = generateTreesMaterial(zip,"trees3.png");
      DRImage* detailedImage = zip.getImage("detail.jpg");
      GLTexture* detailedTexture = new GLTexture(*detailedImage,true);
      delete detailedImage;
      generateIsle(
        glWin,zip,"terrain1.jpg",detailedTexture,16,"terrain1.png",4,
        256,256,32,500,0,128,treesMaterial1,26,&isle[0],&trees[0]
      );
      generateIsle(
        glWin,zip,"terrain2.jpg",detailedTexture,8,"terrain2.png",4,
        256,256,32,1000,500,128,treesMaterial2,24,&isle[1],&trees[1]
      );
      generateIsle(
        glWin,zip,"terrain3.jpg",detailedTexture,16,"terrain3.png",4,
        256,256,12,1000,-500,128,treesMaterial3,4,&isle[2],&trees[2]
      );
    }
    alarmSample = zip.get3DSoundSample("alarm.wav");
    alarmSample->setLooping(true);
    alarmSample->setMinDistance(100);
    alarmSample->setVolume(255);
    motorSample = zip.get3DSoundSample("motor.wav");
    motorSample->setLooping(true);
    motorSample->setMinDistance(50);
    motorSample->setVolume(255);
    fireSample = zip.get3DSoundSample("fire.wav");
    fireSample->setLooping(true);
    fireSample->setMinDistance(20);
    fireSample->setVolume(255);
    waterSample = zip.getSoundSample("water.wav");
    waterSample->setLooping(true);
    windSample = zip.getSoundSample("wind.wav");
    windSample->setLooping(true);
    soundTrack = zip.getMusic("dvorak.mid");
    soundTrack->setVolume(64);
    soundTrack->play();
    sequenceTime = 0;
  } catch(Exception& e) {
    return false;
  }
  return true;
}

GLObjects* IntroScene::cloneMeshes(GLObjects& objs) {
  GLObjects* clone = new GLObjects();
  GLMesh* mesh = (GLMesh*)objs.getFirstObject();
  while(mesh) {
    GLShape& shp = mesh->getShape();
    GLMaterial& mat = mesh->getMaterial();
    clone->addObject(new GLMesh(&shp,&mat));
    mesh = (GLMesh*)objs.getNextObject();
  }
  return clone;
}

GLMaterial* IntroScene::generateTreesMaterial(DRZipFile& zip, char* fileName) {
  DRImage* treesImage;
  treesImage = zip.getImage(fileName);
  GLTexture* treesTexture = new GLTexture(*treesImage);
  delete treesImage;
  GLMaterial* treesMaterial = new GLMaterial();
  treesMaterial->setAmbient(1,1,1);
  treesMaterial->setDiffuse(1,1,1);
  treesMaterial->setDiffuseTexture(treesTexture);
  return treesMaterial;
}

void IntroScene::generateIsle(
  GLWin& glWin, DRZipFile& zip, char* txtName, GLTexture* detail,
  int detailRepeat, char* hghtName, float waterLev, float width, float depth,
  float height, float x, float y, int treesCount, GLMaterial* treesMaterial,
  float minTreesH, GLHeightField** heightField, GLTrees** trees
) {
  DRImage* terrainImage = zip.getImage(txtName);
  GLTexture* terrainTexture = new GLTexture(*terrainImage);
  delete terrainImage;
  GLMaterial* terrainMaterial = new GLMaterial();
  terrainMaterial->setDiffuse(1,1,1);
  terrainMaterial->setDiffuseTexture(terrainTexture);
  terrainMaterial->setGlossTexture(detail);
  DRImage* hfImage = zip.getImage(hghtName);
  *heightField =
    new GLHeightField(hfImage,terrainMaterial,width,depth,height,waterLev,8);
  (*heightField)->move(x,-waterLev,y);
  glWin.getWorld().addObject(*heightField);
  *trees = new GLTrees(treesCount,2,treesMaterial,2,0.1f);
  int counter = 0;
  const int treesSize = 6;
  while(counter < treesCount) {
    float xx = random((int)(width+0.5f))-width*0.5f;
    float yy = random((int)(depth+0.5f))-depth*0.5f;
    float h = (*heightField)->getHeightAtRelative(xx,yy);
    if(h > minTreesH) {
      counter++;
      (*trees)->addTree(
        xx,h,yy,treesSize,treesSize,counter%4,counter<(treesCount>>1)
      );
    }
  }
  (*trees)->setTransparent(true);
  (*trees)->move(x,2.5f,y);
  glWin.getWorld().addObject(*trees);
}

void IntroScene::finalize(GLWin& glWin) {
  glWin.setHelpMode(GLWin::REDUCED_HELP);
  glWin.getWorld().empty();
  if(carrier) carrier = NULL;
  for(int ct = 0; ct < 3; ct++) {
    zero[ct] = NULL;
    wild[ct] = NULL;
    if(isle[ct]) {
      trees[ct] = NULL;
      isle[ct] = NULL;
    }
  }
  for(int ct = 0; ct < 2; ct++) {
    fireEmitter[ct] = NULL;
    smokeEmitter[ct] = NULL;
  }
  if(starBackground) starBackground = NULL;
  if(scrollingParagraph) scrollingParagraph = NULL;
  if(slidingParagraph) slidingParagraph = NULL;
  if(soundTrack) {
    soundTrack->stop();
    delete soundTrack;
    soundTrack = NULL;
  }
  if(alarmSample) {
    delete alarmSample;
    alarmSample = NULL;
  }
  if(fireSample) {
    delete fireSample;
    fireSample = NULL;
  }
  if(motorSample) {
    delete motorSample;
    motorSample = NULL;
  }
  if(waterSound) {
    waterSound->stop();
    delete waterSound;
    delete waterSample;
    waterSound = NULL;
    waterSample = NULL;
  }
  if(windSound) {
    windSound->stop();
    delete windSound;
    delete windSample;
    windSound = NULL;
    windSample = NULL;
  }
}

int IntroScene::update(GLWin& glWin) {
  GLCamera& camera = glWin.getCamera();
  float timeStep = GLWin::getTimeStep();
  sequenceTime += timeStep;
  switch(sequence) {
    default:
    case SCROLLING_TEXTS: {
      if(sequenceTime < 60) {
        starBackground->yaw(timeStep*0.06f);
        starBackground->pitch(timeStep*0.04f);
        slidingParagraph->move(0,timeStep*0.6f,0);
        scrollingParagraph->move(0,timeStep*0.2f,timeStep*0.2f);
      } else {
        camera.pitch(-timeStep*0.2f);
        camera.rotateStanding(-timeStep*0.2f);
        if(scrollingParagraph) {
          scrollingParagraph->move(0,timeStep*0.2f,timeStep*0.2f);
          if(slidingParagraph) {
            slidingParagraph->move(0,timeStep*0.6f,0);
            if(slidingParagraph->isClipped()) {
              glWin.getWorld().removeObject(slidingParagraph);
              slidingParagraph = NULL;
            }
          }
          if(scrollingParagraph->isClipped()) {
            glWin.getWorld().removeObject(scrollingParagraph);
            scrollingParagraph = NULL;
            sun->setVisible(true);
            sun->setColor(0,0,0);
            sea->setVisible(true);
            for(int ct = 0; ct < 3; ct++) {
              isle[ct]->setVisible(true);
              trees[ct]->setVisible(true);
            }
            glWin.getWorld().addBackground(skyBackground);
            glWin.getWorld().enableFog(1000,GLColor(0,0,0));
            glWin.getWorld().setAmbient(0,0,0);
            sequence = SUN_RISE;
          }
        }
      }
      break;
    }
    case SUN_RISE: {
      if(starBackground) {
        float clr = starBackground->getColor().getRed();
        clr -= timeStep*0.05f;
        if(clr < 0) {
          clr = 0;
          glWin.getWorld().removeBackground(starBackground);
          starBackground = NULL;
        } else {
          starBackground->setColor(clr,clr,clr);
        }
        clr = 1-clr;
        skyBackground->setColor(clr,clr,clr);
        sun->setColor(clr,clr,clr);
        glWin.getWorld().setFogColor(GLColor(clr*0.4f,clr*0.4f,clr));
        clr *= 0.5f;
        glWin.getWorld().setAmbient(clr,clr,clr);
      }
      if(camera.getViewDirection().getY() > 0) {
        camera.pitch(timeStep*0.1f);
        camera.rotateStanding(-timeStep*0.1f);
      } else if(sun->isClipped()) {
        camera.getViewDirection().setY(0);
        sequence = SHOW_ISLES;
      }
      break;
    }
    case SHOW_ISLES: {
      if(starBackground) {
        float clr = starBackground->getColor().getRed();
        clr -= timeStep*0.05f;
        if(clr < 0) {
          clr = 0;
          glWin.getWorld().removeBackground(starBackground);
          starBackground = NULL;
        } else {
          starBackground->setColor(clr,clr,clr);
        }
        clr = 1-clr;
        skyBackground->setColor(clr,clr,clr);
        sun->setColor(clr,clr,clr);
        clr *= 0.5f;
        glWin.getWorld().setAmbient(clr,clr,clr);
      } else if(sun->isClipped()) {
        if(soundTrack) {
          soundTrack->stop();
          delete soundTrack;
          soundTrack = NULL;
        }
        waterSound = waterSample->createSound();
        waterSound->play(255);
        waterSound->setVolume(255);
        windSound = windSample->createSound();
        windSound->play(255);
        windSound->setVolume(32);
        sequence = APPROACH_ISLES;
      }
      camera.rotateStanding(-timeStep*0.075f);
      break;
    }
    case APPROACH_ISLES: {
      M3Vector distIsle1 = isle[0]->getPosition();
      distIsle1.subtract(
        camera.getPositionX(),camera.getPositionY(),camera.getPositionZ()
      );
      if(distIsle1.dot(camera.getSideDirection()) < 0) {
        camera.rotateStanding(-timeStep*0.1f);
      } else {
        sequence = APPROACH_ISLES_2;
      }
      break;
    }
    case APPROACH_ISLES_2: {
      M3Vector distIsle1 = isle[0]->getPosition();
      distIsle1.subtract(
        camera.getPositionX(),camera.getPositionY(),camera.getPositionZ()
      );
      float norm2 = distIsle1.norm2();
      float rad2 = isle[0]->getMaxRadius()*isle[0]->getMaxRadius();
      const float maxHeight = 50;
      const float minHeight = 32;
      static float lastHeight = 0;
      if(abs(camera.getPositionY()-lastHeight) > 1) {
        lastHeight = camera.getPositionY();
        int windVol = (int)(lastHeight*(255.0f/30));
        if(windVol > 255) windVol = 255;
        int waterVol = 255-windVol;
        windSound->setVolume(windVol);
        waterSound->setVolume(waterVol);
      }
      if(norm2 > rad2*.125f) {
        camera.moveForward(timeStep*20);
        float h = camera.getPositionY()+1.5f*timeStep;
        if(h < maxHeight) camera.setPositionY(h);
      } else {
        distIsle1.setY(minHeight-camera.getPositionY());
        distIsle1.normalize().scale(timeStep*4);
        camera.move(distIsle1);
        camera.rotateStanding(timeStep*0.15f);
        M3Vector sunDir = sun->getDirection();
        if(sunDir.dot(camera.getSideDirection()) < 0) {
          sequence = WATCH_TREES;
          sequenceTime = 0;
        }
      }
      break;
    }
    case WATCH_TREES: {
      if(sequenceTime >= 5.0f) {
        for(int ct = 0; ct < 2; ct++) {
          fireEmitter[ct]->setVisible(true);
          smokeEmitter[ct]->setVisible(true);
        }
        carrier->setVisible(true);
        for(int ct = 0; ct < 3; ct++) {
          zero[ct]->setVisible(true);
          wild[ct]->setVisible(true);
        }
        GLSoundSource* soundSource;
        soundSource =
          new GLSoundSource(alarmSample,carrier);
        glWin.getWorld().addSoundSource(soundSource);
        soundSource =
          new GLSoundSource(fireSample,fireEmitter[0]);
        glWin.getWorld().addSoundSource(soundSource);
        soundSource =
          new GLSoundSource(fireSample,fireEmitter[1]);
        glWin.getWorld().addSoundSource(soundSource);
        for(int ct = 0; ct < 3; ct++) {
          soundSource =
            new GLSoundSource(motorSample,zero[ct]);
          glWin.getWorld().addSoundSource(soundSource);
          soundSource =
            new GLSoundSource(motorSample,wild[ct]);
          glWin.getWorld().addSoundSource(soundSource);
        }
        camera.rotateStanding(timeStep*0.1f);
        sequence = WATCH_CARRIER;
      }
      break;
    }
    case WATCH_CARRIER: {
      const float airplaneSpeed = 50;
      for(int ct = 0; ct < 3; ct++) {
        zero[ct]->move(airplaneSpeed*timeStep,0,0);
        if(zero[ct]->getPositionX() > 3000) zero[ct]->setPositionX(-1500);
        wild[ct]->move(airplaneSpeed*timeStep,0,0);
        if(wild[ct]->getPositionX() > 3000) wild[ct]->setPositionX(-1500);
      }
      for(int ct = 0; ct < 2; ct++) {
        fireEmitter[ct]->move(airplaneSpeed*timeStep,0,0);
        if(fireEmitter[ct]->getPositionX() > 3000)
          fireEmitter[ct]->setPositionX(-1500);
        smokeEmitter[ct]->move(airplaneSpeed*timeStep,0,0);
        if(smokeEmitter[ct]->getPositionX() > 3000)
          smokeEmitter[ct]->setPositionX(-1500);
      }
      if(camera.getViewDirectionX() < 0) {
        camera.move(0,timeStep*0.25f,0);
        camera.rotateStanding(timeStep*0.2f);
        sequenceTime = 0;
      } else if(sequenceTime >= 5.0f) {
        sequence = FLY_THROUGH;
        glWin.setHelpMode(GLWin::USER_HELP);
      }
      break;
    }
    case FLY_THROUGH: {
      const float airplaneSpeed = 50;
      for(int ct = 0; ct < 3; ct++) {
        zero[ct]->move(airplaneSpeed*timeStep,0,0);
        if(zero[ct]->getPositionX() > 3000) zero[ct]->setPositionX(-1500);
        wild[ct]->move(airplaneSpeed*timeStep,0,0);
        if(wild[ct]->getPositionX() > 3000) wild[ct]->setPositionX(-1500);
      }
      for(int ct = 0; ct < 2; ct++) {
        fireEmitter[ct]->move(airplaneSpeed*timeStep,0,0);
        if(fireEmitter[ct]->getPositionX() > 3000)
          fireEmitter[ct]->setPositionX(-1500);
        smokeEmitter[ct]->move(airplaneSpeed*timeStep,0,0);
        if(smokeEmitter[ct]->getPositionX() > 3000)
          smokeEmitter[ct]->setPositionX(-1500);
      }
      static float speed = 0;
      camera.moveForward(speed*timeStep);
      if(glWin.isKeyPressed(VK_UP)) {
        speed += 15*timeStep;
        if(speed > 55) speed = 55;
      }
      if(glWin.isKeyPressed(VK_DOWN)) {
        speed -= 15*timeStep;
        if(speed < 0) speed = 0;
      }
      if(glWin.isKeyPressed(VK_LEFT)) {
        camera.roll(-0.2f*timeStep);
      }
      if(glWin.isKeyPressed(VK_RIGHT)) {
        camera.roll(0.2f*timeStep);
      }
      int dx = glWin.getMouseDX();
      int dy = glWin.getMouseDY();
      float changeStep = 0.15f*timeStep;
      if(dx != 0) camera.yaw(-dx*changeStep);
      if(dy != 0) camera.pitch(-dy*changeStep);
      for(int ct = 0; ct < 3; ct++) {
        if(isle[ct]->bounds(camera.getPosition())) {
          float h =
            isle[ct]->getHeightAtAbsolute(
              camera.getPositionX(),camera.getPositionZ()
            )+5.5f;
          if(camera.getPositionY() < h) {
            camera.setPositionY(h);
            break;
          }
        }
      }
      if(camera.getPositionY() < 5.5f)
        camera.setPositionY(5.5f);
      else if(camera.getPositionY() > 2500)
        camera.setPositionY(2500);
      static float lastHeight = 0;
      if(camera.getPositionY() != lastHeight) {
        lastHeight = camera.getPositionY();
        int windVol = (int)(lastHeight/30*255);
        if(windVol > 255) windVol = 255;
        int waterVol = 255-windVol;
        windSound->setVolume(windVol);
        waterSound->setVolume(waterVol);
      }
      break;
    }
  }

  return INTRO_SCENE;
}

char* IntroScene::getHelpLine(int line) {
  switch(line) {
    case 13: return "\"Zeke On Your Six!\"";
    case 12: return "Full Interactive Version";
    case 11: return "available in October 2002";
    case 10: return "at http://simulaxion.sf.net";
    case 9: return " ";
    case 8: return "Contact boselli@uno.it";
    case 7: return "for more info";
    case 6: return " ";
    case 5: return "[MOUSE] Change Direction";
    case 4: return "[UP   ] Increase Speed";
    case 3: return "[DOWN ] Reduce Speed";
    case 2: return "[LEFT ] Roll Left";
    case 1: return "[RIGHT] Roll Right";
    case 0: return "[F1]    Help";
    default: return NULL;
  }
}

//
// GameScene
//
class GameScene: public BasicScene {
private:
  FMSound *waterSound, *windSound;
  FMSample *waterSample, *windSample;
  FM3DSample *alarmSample, *fireSample;
  FM3DSample *motorSample;
  GLBackground *skyBackground;
  GLFlatTerrain* sea;
  GLSunLight* sun;
  GLHeightField *isle[3];
  GLTrees *trees[3];
  GLObjects *carrier, *zero[3], *wild[3];
  GLEmitter *fireEmitter[2], *smokeEmitter[2];
protected:
  void generateIsle(
    GLWin& glWin, DRZipFile& zip, char* txtName, GLTexture* detail,
    int detailRepeat, char* hghtName, float waterLev, float width, float depth,
    float height, float x, float y, int treesCount, GLMaterial* treesMaterial,
    float minTreesH, GLHeightField** isle, GLTrees** trees
  );
  GLMaterial* generateTreesMaterial(DRZipFile& zip, char* fileName);
  GLObjects* cloneMeshes(GLObjects& objs);
public:
  GameScene();
  bool initialize(GLWin& glWin);
  void finalize(GLWin& glWin);
  int update(GLWin& glWin);
  char* getHelpLine(int line);
};

GameScene::GameScene():
  waterSound(NULL), waterSample(NULL),
  alarmSample(NULL), fireSample(NULL),
  motorSample(NULL),
  windSound(NULL), windSample(NULL),
  skyBackground(NULL),
  sun(NULL), sea(NULL),
  carrier(NULL)
{
  for(int ct = 0; ct < 3; ct++) {
    isle[ct] = NULL;
    trees[ct] = NULL;
    zero[ct] = NULL;
    wild[ct] = NULL;
  }
  for(int ct = 0; ct < 2; ct++) {
    fireEmitter[ct] = NULL;
    smokeEmitter[ct] = NULL;
  }
}

bool GameScene::initialize(GLWin& glWin) {
  glWin.setHelpMode(GLWin::REDUCED_HELP);
  glWin.setPerspective(60,0.5f,3000);
  glWin.getCamera().setPosition(0,10,0);
  glWin.getWorld().setAmbient(1,1,1);
  try {
    DRZipFile zip("ZekeOnYourSix!.dat");
    /// background
    {
      const int texturesCount = 5;
      DRImage* skyboxImages[texturesCount] = {
        zip.getImage("SkyboxTop.jpg"), zip.getImage("SkyboxLeft.jpg"),
        zip.getImage("SkyboxFront.jpg"), zip.getImage("SkyboxRight.jpg"),
        zip.getImage("SkyboxBack.jpg"),
      };
      GLTexture* skyboxTextures[texturesCount];
      for(int ct = 0; ct < texturesCount; ct++)
        skyboxTextures[ct] = new GLTexture(*skyboxImages[ct]);
      for(int ct = 0; ct < texturesCount; ct++) delete skyboxImages[ct];
      skyBackground = new GLMirroredSkybox(skyboxTextures);
      skyBackground->setColor(0,0,0);
      skyBackground->rotateStanding(3.14159);
    }
    /// sun
    {
      DRImage* sunImage = zip.getImage("light.jpg");
      GLTexture* sunTexture = new GLTexture(*sunImage);
      delete sunImage;
      DRImage* flareImage = zip.getImage("lensflares.png");
      GLTexture* flareTexture = new GLTexture(*flareImage);
      delete flareImage;
      M3Vector direc(0,0.413793f,-0.910345);
      sun = new GLSunLight(
        sunTexture,0.25f,direc,flareTexture,7,0.2f
      );
      sun->setVisible(true);
      sun->setColor(1,1,1);
      glWin.getWorld().setSun(sun);
    }
    /// sea
    {
      const int imagesCount = 16;
      DRImage* waterImages[imagesCount] = {
        zip.getImage("water00.jpg"),
        zip.getImage("water01.jpg"),
        zip.getImage("water02.jpg"),
        zip.getImage("water03.jpg"),
        zip.getImage("water04.jpg"),
        zip.getImage("water05.jpg"),
        zip.getImage("water06.jpg"),
        zip.getImage("water07.jpg"),
        zip.getImage("water08.jpg"),
        zip.getImage("water09.jpg"),
        zip.getImage("water10.jpg"),
        zip.getImage("water11.jpg"),
        zip.getImage("water12.jpg"),
        zip.getImage("water13.jpg"),
        zip.getImage("water14.jpg"),
        zip.getImage("water15.jpg"),
      };
      GLAnimatedTexture* waterTexture =
        new GLAnimatedTexture(imagesCount,waterImages,1.5f,true);
      for(int ct = 0; ct < imagesCount; ct++) delete waterImages[ct];
      GLMaterial* terrainMaterial = new GLMaterial();
      terrainMaterial->setAmbient(0.7f,0.7f,0.7f);
      terrainMaterial->setDiffuse(0,0.5f,0.9f);
      terrainMaterial->setDiffuseTexture(waterTexture);
      sea = new GLFlatTerrain(terrainMaterial,3000,200);
      sea->setVisible(false);
      sea->setReflective(true);
      glWin.getWorld().setFlatTerrain(sea);
    }
    /// 3DS
    {
      carrier = zip.getMeshes("carrier.3ds");
      carrier->move(600,-2,300);
      carrier->pitch(-3.1415/2);
      carrier->rotateStanding(-3.1415/2);
      carrier->setVisible(false);
      glWin.getWorld().addObject(carrier);
      zero[0] = zip.getMeshes("zero.3ds");
      zero[0]->move(-300,15,300);
      zero[0]->pitch(-3.1415/2);
      zero[0]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(zero[0]);
      zero[1] = cloneMeshes(*zero[0]);
      zero[1]->move(-270,55,320);
      zero[1]->pitch(-3.1415/2);
      zero[1]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(zero[1]);
      zero[2] = cloneMeshes(*zero[0]);
      zero[2]->move(-280,50,310);
      zero[2]->pitch(-3.1415/2);
      zero[2]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(zero[2]);
      wild[0] = zip.getMeshes("wildcat.3ds");
      wild[0]->move(-500,30,310);
      wild[0]->pitch(-3.1415/2);
      wild[0]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(wild[0]);
      wild[1] = cloneMeshes(*wild[0]);
      wild[1]->move(-510,20,320);
      wild[1]->pitch(-3.1415/2);
      wild[1]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(wild[1]);
      wild[2] = cloneMeshes(*wild[0]);
      wild[2]->move(-525,65,330);
      wild[2]->pitch(-3.1415/2);
      wild[2]->rotateStanding(-3.1415/2);
      glWin.getWorld().addObject(wild[2]);
    }
    /// emitters
    {
      DRImage* fireImage = zip.getImage("fire.png");
      fireImage->convertGray2111A();
      GLTexture* fireTexture = new GLTexture(*fireImage);
      DRImage* smokeImage = zip.getImage("smoke.png");
      smokeImage->convertGray2RGBA();
      GLTexture* smokeTexture = new GLTexture(*smokeImage);
      for(int ct = 0; ct < 2; ct++) {
        fireEmitter[ct] = new GLEmitter(10,1,15,true);
        fireEmitter[ct]->setParticleTexture(
          fireTexture,fireImage->getHasAlpha()
        );
        fireEmitter[ct]->setParticleVelocity(M3Vector(-3,2,0),0.5f);
        fireEmitter[ct]->setColor(GLColor(1,1,0,1),GLColor(1,0,0,0));
        fireEmitter[ct]->setSize(2,.5);
        fireEmitter[ct]->setGravity(M3Vector(),M3Vector());
        fireEmitter[ct]->setVisible(false);
        glWin.getWorld().addObject(fireEmitter[ct]);
        smokeEmitter[ct] = new GLEmitter(20,4,30,true);
        smokeEmitter[ct]->setParticleTexture(
          smokeTexture,smokeImage->getHasAlpha()
        );
        smokeEmitter[ct]->setParticleVelocity(M3Vector(-4,1,0),0.5f);
        smokeEmitter[ct]->setColor(GLColor(0,0,0,1),GLColor(0,0,0,0));
        smokeEmitter[ct]->setSize(.25,2);
        smokeEmitter[ct]->setGravity(M3Vector(),M3Vector(0,-4,0));
        smokeEmitter[ct]->setVisible(false);
        glWin.getWorld().addObject(smokeEmitter[ct]);
      }
      delete fireImage;
      delete smokeImage;
      fireEmitter[0]->move(-514,20,320);
      smokeEmitter[0]->move(-514,20,320);
      fireEmitter[1]->move(-304,15,300);
      smokeEmitter[1]->move(-304,15,300);
    }
    /// isles
    {
      GLMaterial* treesMaterial1 = generateTreesMaterial(zip,"trees1.png");
      GLMaterial* treesMaterial2 = generateTreesMaterial(zip,"trees2.png");
      GLMaterial* treesMaterial3 = generateTreesMaterial(zip,"trees3.png");
      DRImage* detailedImage = zip.getImage("detail.jpg");
      GLTexture* detailedTexture = new GLTexture(*detailedImage,true);
      delete detailedImage;
      generateIsle(
        glWin,zip,"terrain1.jpg",detailedTexture,16,"terrain1.png",4,
        256,256,32,500,0,128,treesMaterial1,26,&isle[0],&trees[0]
      );
      generateIsle(
        glWin,zip,"terrain2.jpg",detailedTexture,8,"terrain2.png",4,
        256,256,32,1000,500,128,treesMaterial2,24,&isle[1],&trees[1]
      );
      generateIsle(
        glWin,zip,"terrain3.jpg",detailedTexture,16,"terrain3.png",4,
        256,256,12,1000,-500,128,treesMaterial3,4,&isle[2],&trees[2]
      );
    }
    alarmSample = zip.get3DSoundSample("alarm.wav");
    alarmSample->setLooping(true);
    alarmSample->setMinDistance(100);
    alarmSample->setVolume(255);
    motorSample = zip.get3DSoundSample("motor.wav");
    motorSample->setLooping(true);
    motorSample->setMinDistance(50);
    motorSample->setVolume(255);
    fireSample = zip.get3DSoundSample("fire.wav");
    fireSample->setLooping(true);
    fireSample->setMinDistance(20);
    fireSample->setVolume(255);
    waterSample = zip.getSoundSample("water.wav");
    waterSample->setLooping(true);
    windSample = zip.getSoundSample("wind.wav");
    windSample->setLooping(true);
  } catch(Exception& e) {
    return false;
  }
  sun->setVisible(true);
  sun->setColor(1,1,0.5f);
  sea->setVisible(true);
  for(int ct = 0; ct < 3; ct++) {
    isle[ct]->setVisible(true);
    trees[ct]->setVisible(true);
  }
  glWin.getWorld().setBackground(skyBackground);
  skyBackground->setColor(1,1,1);
  glWin.getWorld().enableFog(1000,GLColor(0.4f,0.4f,1));
  glWin.getWorld().setAmbient(0.5f,0.5f,0.5f);
  waterSound = waterSample->createSound();
  waterSound->play(255);
  waterSound->setVolume(255);
  windSound = windSample->createSound();
  windSound->play(255);
  windSound->setVolume(32);
  for(int ct = 0; ct < 2; ct++) {
    fireEmitter[ct]->setVisible(true);
    smokeEmitter[ct]->setVisible(true);
  }
  carrier->setVisible(true);
  for(int ct = 0; ct < 3; ct++) {
    zero[ct]->setVisible(true);
    wild[ct]->setVisible(true);
  }
  GLSoundSource* soundSource;
  soundSource =
    new GLSoundSource(alarmSample,carrier);
  glWin.getWorld().addSoundSource(soundSource);
  soundSource =
    new GLSoundSource(fireSample,fireEmitter[0]);
  glWin.getWorld().addSoundSource(soundSource);
  soundSource =
    new GLSoundSource(fireSample,fireEmitter[1]);
  glWin.getWorld().addSoundSource(soundSource);
  for(int ct = 0; ct < 3; ct++) {
    soundSource =
      new GLSoundSource(motorSample,zero[ct]);
    glWin.getWorld().addSoundSource(soundSource);
    soundSource =
      new GLSoundSource(motorSample,wild[ct]);
    glWin.getWorld().addSoundSource(soundSource);
  }
  return true;
}

GLObjects* GameScene::cloneMeshes(GLObjects& objs) {
  GLObjects* clone = new GLObjects();
  GLMesh* mesh = (GLMesh*)objs.getFirstObject();
  while(mesh) {
    GLShape& shp = mesh->getShape();
    GLMaterial& mat = mesh->getMaterial();
    clone->addObject(new GLMesh(&shp,&mat));
    mesh = (GLMesh*)objs.getNextObject();
  }
  return clone;
}

GLMaterial* GameScene::generateTreesMaterial(DRZipFile& zip, char* fileName) {
  DRImage* treesImage;
  treesImage = zip.getImage(fileName);
  GLTexture* treesTexture = new GLTexture(*treesImage);
  delete treesImage;
  GLMaterial* treesMaterial = new GLMaterial();
  treesMaterial->setAmbient(1,1,1);
  treesMaterial->setDiffuse(1,1,1);
  treesMaterial->setDiffuseTexture(treesTexture);
  return treesMaterial;
}

void GameScene::generateIsle(
  GLWin& glWin, DRZipFile& zip, char* txtName, GLTexture* detail,
  int detailRepeat, char* hghtName, float waterLev, float width, float depth,
  float height, float x, float y, int treesCount, GLMaterial* treesMaterial,
  float minTreesH, GLHeightField** heightField, GLTrees** trees
) {
  DRImage* terrainImage = zip.getImage(txtName);
  GLTexture* terrainTexture = new GLTexture(*terrainImage);
  delete terrainImage;
  GLMaterial* terrainMaterial = new GLMaterial();
  terrainMaterial->setDiffuse(1,1,1);
  terrainMaterial->setDiffuseTexture(terrainTexture);
  terrainMaterial->setGlossTexture(detail);
  DRImage* hfImage = zip.getImage(hghtName);
  *heightField =
    new GLHeightField(hfImage,terrainMaterial,width,depth,height,waterLev,8);
  (*heightField)->move(x,-waterLev,y);
  glWin.getWorld().addObject(*heightField);
  *trees = new GLTrees(treesCount,2,treesMaterial,2,0.1f);
  int counter = 0;
  const int treesSize = 6;
  while(counter < treesCount) {
    float xx = random((int)(width+0.5f))-width*0.5f;
    float yy = random((int)(depth+0.5f))-depth*0.5f;
    float h = (*heightField)->getHeightAtRelative(xx,yy);
    if(h > minTreesH) {
      counter++;
      (*trees)->addTree(
        xx,h,yy,treesSize,treesSize,counter%4,counter<(treesCount>>1)
      );
    }
  }
  (*trees)->setTransparent(true);
  (*trees)->move(x,2.5f,y);
  glWin.getWorld().addObject(*trees);
}

void GameScene::finalize(GLWin& glWin) {
  glWin.setHelpMode(GLWin::REDUCED_HELP);
  glWin.getWorld().empty();
  if(carrier) carrier = NULL;
  for(int ct = 0; ct < 3; ct++) {
    zero[ct] = NULL;
    wild[ct] = NULL;
    if(isle[ct]) {
      trees[ct] = NULL;
      isle[ct] = NULL;
    }
  }
  for(int ct = 0; ct < 2; ct++) {
    fireEmitter[ct] = NULL;
    smokeEmitter[ct] = NULL;
  }
  if(alarmSample) {
    delete alarmSample;
    alarmSample = NULL;
  }
  if(fireSample) {
    delete fireSample;
    fireSample = NULL;
  }
  if(motorSample) {
    delete motorSample;
    motorSample = NULL;
  }
  if(waterSound) {
    waterSound->stop();
    delete waterSound;
    delete waterSample;
    waterSound = NULL;
    waterSample = NULL;
  }
  if(windSound) {
    windSound->stop();
    delete windSound;
    delete windSample;
    windSound = NULL;
    windSample = NULL;
  }
}

int GameScene::update(GLWin& glWin) {
  GLCamera& camera = glWin.getCamera();
  if(glWin.isKeyPressed('1')) {
    glWin.keyUp('1');
    return INTRO_SCENE;
  }
  float timeStep = GLWin::getTimeStep();
  const float airplaneSpeed = 50;
  for(int ct = 0; ct < 3; ct++) {
    zero[ct]->move(airplaneSpeed*timeStep,0,0);
    if(zero[ct]->getPositionX() > 3000) zero[ct]->setPositionX(-1500);
    wild[ct]->move(airplaneSpeed*timeStep,0,0);
    if(wild[ct]->getPositionX() > 3000) wild[ct]->setPositionX(-1500);
  }
  for(int ct = 0; ct < 2; ct++) {
    fireEmitter[ct]->move(airplaneSpeed*timeStep,0,0);
    if(fireEmitter[ct]->getPositionX() > 3000)
      fireEmitter[ct]->setPositionX(-1500);
    smokeEmitter[ct]->move(airplaneSpeed*timeStep,0,0);
    if(smokeEmitter[ct]->getPositionX() > 3000)
      smokeEmitter[ct]->setPositionX(-1500);
  }
  static float speed = 0;
  camera.moveForward(speed*timeStep);
  if(glWin.isKeyPressed(VK_UP)) {
    speed += 15*timeStep;
    if(speed > 55) speed = 55;
  }
  if(glWin.isKeyPressed(VK_DOWN)) {
    speed -= 15*timeStep;
    if(speed < 0) speed = 0;
  }
  if(glWin.isKeyPressed(VK_LEFT)) {
    camera.roll(-0.2f*timeStep);
  }
  if(glWin.isKeyPressed(VK_RIGHT)) {
    camera.roll(0.2f*timeStep);
  }
  int dx = glWin.getMouseDX();
  int dy = glWin.getMouseDY();
  float changeStep = 0.15f*timeStep;
  if(dx != 0) camera.yaw(-dx*changeStep);
  if(dy != 0) camera.pitch(-dy*changeStep);
  for(int ct = 0; ct < 3; ct++) {
    if(isle[ct]->bounds(camera.getPosition())) {
      float h =
        isle[ct]->getHeightAtAbsolute(
          camera.getPositionX(),camera.getPositionZ()
        )+5.5f;
      if(camera.getPositionY() < h) {
        camera.setPositionY(h);
        break;
      }
    }
  }
  if(camera.getPositionY() < 5.5f)
    camera.setPositionY(5.5f);
  else if(camera.getPositionY() > 2500)
    camera.setPositionY(2500);
  static float lastHeight = 0;
  if(camera.getPositionY() != lastHeight) {
    lastHeight = camera.getPositionY();
    int windVol = (int)(lastHeight/30*255);
    if(windVol > 255) windVol = 255;
    int waterVol = 255-windVol;
    windSound->setVolume(windVol);
    waterSound->setVolume(waterVol);
  }
  return GAME_SCENE;
}

char* GameScene::getHelpLine(int line) {
  switch(line) {
    case 14: return "\"Zeke On Your Six!\"";
    case 13: return "Full Interactive Version";
    case 12: return "available in October 2002";
    case 11: return "at http://simulaxion.sf.net";
    case 10: return " ";
    case 9: return "Contact boselli@uno.it";
    case 8: return "for more info";
    case 7: return " ";
    case 6: return "[MOUSE] Change Direction";
    case 5: return "[UP   ] Increase Speed";
    case 4: return "[DOWN ] Reduce Speed";
    case 3: return "[LEFT ] Roll Left";
    case 2: return "[RIGHT] Roll Right";
    case 1: return "[1]     Replay Intro";
    case 0: return "[F1]    Help";
    default: return NULL;
  }
}

//
// ZekeOnYourSix
//
class ZekeOnYourSix: public GLApp {
protected:
  char* getTitle() {return "\"Zeke On Your Six!\" (Four Elements III Demo)";}
  char* getHelpLine(int line);
  bool initializeScene();
  void finalizeScene();
};

char* ZekeOnYourSix::getHelpLine(int line) {
  return ((BasicScene*)getCurrentScene())->getHelpLine(line);
}

bool ZekeOnYourSix::initializeScene() {
  try {
    DRZipFile zip("ZekeOnYourSix!.dat");
    /// font
    {
      DRImage* fontImage = zip.getImage("font.png");
      fontImage->addAlpha(*fontImage);
      GLTexture* fontTexture = new GLTexture(*fontImage);
      setTextOverlayFont(new GLTexturedFont(16,16,10,fontTexture));
      delete fontImage;
    }
  } catch(Exception& e) {
    showError("Error Loading DAT File");
  }
  deleteScenes();
  BasicScene* scenes[] = {
    new SplashScene(),
    new IntroScene(),
    new GameScene(),
    NULL,
  };
  addScenes((GLScene**)scenes);
  setCurrentScene(BasicScene::SPLASH_SCENE);
  return true;
}

void ZekeOnYourSix::finalizeScene() {
  finalizeCurrent();
}

//
// GLWin
//

GLWin* GLWin::newInstance(char* initStr) throw(Exception) {
  return new ZekeOnYourSix();
}
