/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "dataread.h"
#include "glmesh.h"
#include "fmsound.h"

#include "datasets.cpp"

#include "jpeg/jerror.h"

//
// Jpeg Aux (common)
//
METHODDEF(void) my_output_message(j_common_ptr cinfo) {
  char buffer[JMSG_LENGTH_MAX];
  (*cinfo->err->format_message)(cinfo, buffer);
  Exception::showError(buffer,"JPEG");
}

//
// Jpeg Aux (ZIP)
//
typedef struct {
  struct jpeg_source_mgr pub;
  DRZipFile* zipFile;
  char* zippedFileName;
} my_source_mgr_zip;

METHODDEF(void) init_source_zip(j_decompress_ptr cinfo) {
  my_source_mgr_zip* src = (my_source_mgr_zip*)cinfo->src;
  src->zipFile->createZipBuffer(src->zippedFileName);
}

METHODDEF(boolean) fill_input_buffer_zip(j_decompress_ptr cinfo) {
  my_source_mgr_zip* src = (my_source_mgr_zip*)cinfo->src;
  src->pub.next_input_byte = src->zipFile->getZipBuffer();
  src->pub.bytes_in_buffer = (size_t)src->zipFile->getZipBufferLength();
  return TRUE;
}

METHODDEF(void) skip_input_data_zip(j_decompress_ptr cinfo, long num_bytes) {
  my_source_mgr_zip* src = (my_source_mgr_zip*)cinfo->src;
  if(num_bytes > 0) {
    while(num_bytes > (long) src->pub.bytes_in_buffer) {
      num_bytes -= (long) src->pub.bytes_in_buffer;
      fill_input_buffer_zip(cinfo);
    }
    src->pub.next_input_byte += (size_t)num_bytes;
    src->pub.bytes_in_buffer -= (size_t)num_bytes;
  }
}

METHODDEF(void) term_source_zip(j_decompress_ptr cinfo) {
  my_source_mgr_zip* src = (my_source_mgr_zip*)cinfo->src;
  src->zipFile->deleteZipBuffer();
}

//
// Jpeg Aux (Resource)
//
typedef struct {
  struct jpeg_source_mgr pub;
  DRResource* resource;
} my_source_mgr_res;

METHODDEF(void) init_source_res(j_decompress_ptr cinfo) {
  cinfo;
}

METHODDEF(boolean) fill_input_buffer_res(j_decompress_ptr cinfo) {
  my_source_mgr_res* src = (my_source_mgr_res*)cinfo->src;
  try {
    src->pub.next_input_byte = src->resource->getPointer();
  } catch(Exception& e) {
    return FALSE;
  }
  src->pub.bytes_in_buffer = src->resource->getSize();
  return TRUE;
}

METHODDEF(void) skip_input_data_res(j_decompress_ptr cinfo, long num_bytes) {
  my_source_mgr_res* src = (my_source_mgr_res*)cinfo->src;
  if(num_bytes > 0) {
    src->pub.next_input_byte += (size_t)num_bytes;
    src->pub.bytes_in_buffer -= (size_t)num_bytes;
  }
}

METHODDEF(void) term_source_res(j_decompress_ptr cinfo) {
	cinfo;
}

//
// Jpeg Aux (File)
//
#define MAX_FILE_BUFFER 4096

typedef struct {
  struct jpeg_source_mgr pub;
  DRFile* file;
  unsigned char buffer[MAX_FILE_BUFFER];
  boolean start_of_file;
} my_source_mgr_fil;

METHODDEF(void) init_source_fil(j_decompress_ptr cinfo) {
  my_source_mgr_fil* src = (my_source_mgr_fil*)cinfo->src;
  src->file->openFile();
  src->start_of_file = TRUE;
}

METHODDEF(boolean) fill_input_buffer_fil(j_decompress_ptr cinfo) {
  my_source_mgr_fil* src = (my_source_mgr_fil*)cinfo->src;
  DWORD nbytes;
	ReadFile(
  	src->file->getFileHandle(),src->buffer,MAX_FILE_BUFFER,&nbytes,NULL
  );
  if(nbytes <= 0) {
    if (src->start_of_file)
      ERREXIT(cinfo, JERR_INPUT_EMPTY);
    WARNMS(cinfo, JWRN_JPEG_EOF);
    src->buffer[0] = (unsigned char)0xFF;
    src->buffer[1] = (unsigned char)JPEG_EOI;
    nbytes = 2;
  }
  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = nbytes;
  src->start_of_file = FALSE;
  return TRUE;
}

METHODDEF(void) skip_input_data_fil(j_decompress_ptr cinfo, long num_bytes) {
  my_source_mgr_fil* src = (my_source_mgr_fil*)cinfo->src;
  if(num_bytes > 0) {
    while(num_bytes > (long)src->pub.bytes_in_buffer) {
      num_bytes -= (long) src->pub.bytes_in_buffer;
      fill_input_buffer_fil(cinfo);
    }
    src->pub.next_input_byte += (size_t)num_bytes;
    src->pub.bytes_in_buffer -= (size_t)num_bytes;
  }
}

METHODDEF(void) term_source_fil(j_decompress_ptr cinfo) {
  my_source_mgr_fil* src = (my_source_mgr_fil*)cinfo->src;
  src->file->closeFile();
}

//
// DRJpegReader
//
DRJpegReader::DRJpegReader() {
  cinfo.err = jpeg_std_error(&jerr);
  jerr.output_message = my_output_message;
}

DRJpegImage* DRJpegReader::readImage(
  jpeg_decompress_struct* cinfo, jpeg_source_mgr* src,
  void (*init_source)(j_decompress_ptr cinfo),
  unsigned char (*fill_input_buffer)(j_decompress_ptr cinfo),
  void (*skip_input_data)(j_decompress_ptr cinfo,long),
  void (*term_source)(j_decompress_ptr cinfo),
  int alphaColor, int tolerance
) {
  DRJpegImage* jpegImage = new DRJpegImage();
  jpegImage->setHasAlpha(alphaColor >= 0);
  jpeg_create_decompress(cinfo);
  cinfo->src = src;
  src->init_source = init_source;
  src->fill_input_buffer = fill_input_buffer;
  src->skip_input_data = skip_input_data;
  src->resync_to_restart = jpeg_resync_to_restart;
  src->term_source = term_source;
  src->bytes_in_buffer = 0;
  src->next_input_byte = NULL;
  jpeg_read_header(cinfo,TRUE);
  jpeg_start_decompress(cinfo);
  jpegImage->setWidth(cinfo->output_width);
  jpegImage->setHeight(cinfo->output_height);
  int components = cinfo->output_components;
  if(jpegImage->getHasAlpha()) components++;
  int row_stride = cinfo->output_width*components;
  unsigned char** jpegBuffer = new unsigned char*[cinfo->output_height];
  unsigned char* buffer = new unsigned char[row_stride*cinfo->output_height];
  for(unsigned int ct = 0; ct < cinfo->output_height; ct++)
    jpegBuffer[ct] = &buffer[ct*row_stride];
  char alphaRed, alphaGreen, alphaBlue;
  if(jpegImage->getHasAlpha()) {
    alphaRed = alphaColor>>16;
    alphaGreen = (alphaColor>>8)&0xff;
    alphaBlue = alphaColor&0xff;
  }
  while(cinfo->output_scanline < cinfo->output_height) {
    int line = cinfo->output_scanline;
    jpeg_read_scanlines(cinfo,&jpegBuffer[line],1);
    if(jpegImage->getHasAlpha()) {
      int ct4 = (jpegImage->getWidth()-1)*4;
      int ct3 = (jpegImage->getWidth()-1)*3;
      for(; ct3 >= 0; ct3 -= 3, ct4 -= 4) {
        unsigned char r = jpegBuffer[line][ct3  ];
        unsigned char g = jpegBuffer[line][ct3+1];
        unsigned char b = jpegBuffer[line][ct3+2];
        jpegBuffer[line][ct4  ] = r;
        jpegBuffer[line][ct4+1] = g;
        jpegBuffer[line][ct4+2] = b;
        jpegBuffer[line][ct4+3] = (
          alphaRed-tolerance   <= r && r <= alphaRed+tolerance   &&
          alphaGreen-tolerance <= g && g <= alphaGreen+tolerance &&
          alphaBlue-tolerance  <= b && b <= alphaBlue+tolerance
        )? 0x00: 0xff;
      }
    }
  }
  delete jpegBuffer;
  jpegImage->setBuffer(buffer);
  jpeg_finish_decompress(cinfo);
  jpeg_destroy_decompress(cinfo);
  return jpegImage;
}

//
// DRPngReader
//
DRPngReader::DRPngReader(): readPointer(NULL) {
}

void DRPngReader::pngWarning(png_structp png_ptr, png_const_charp message) {
  png_ptr;
  Exception::showWarning(message,"PNG");
}

void DRPngReader::pngError(png_structp png_ptr, png_const_charp message) {
  png_ptr;
  Exception::showError(message,"PNG");
}

void DRPngReader::pngReadData(
  png_structp png_ptr, png_bytep data, png_size_t length
) {
  DRPngReader* pr = (DRPngReader*)png_get_io_ptr(png_ptr);
  CopyMemory(data,pr->getReadPointer(),length);
  pr->moveReadPointer(length);
}

DRPngImage* DRPngReader::readImage(
) throw(Exception) {
  png_structp read_ptr = png_create_read_struct(
    PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
    (png_error_ptr)pngError, (png_error_ptr)pngWarning
  );
  if(!read_ptr) Exception::throws("png_create_read_struct");
    png_infop read_info_ptr = png_create_info_struct(read_ptr);
  if(!read_info_ptr) {
    png_destroy_read_struct(&read_ptr,(png_infopp)NULL,(png_infopp)NULL);
    Exception::throws("png_create_info_struct (read_info_ptr)");
  }
  png_infop end_info_ptr = png_create_info_struct(read_ptr);
  if(!end_info_ptr) {
    png_destroy_read_struct(&read_ptr,&read_info_ptr,(png_infopp)NULL);
    Exception::throws("png_create_info_struct (end_info_ptr)");
  }
  if(setjmp(read_ptr->jmpbuf)) {
    png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
    Exception::throws("PNG Read Error");
  }
  png_set_read_fn(read_ptr, (png_voidp)this, (png_rw_ptr)pngReadData);
  png_read_info(read_ptr, read_info_ptr);
  unsigned long width, height;
  int bits, mode;
    png_get_IHDR(
    read_ptr, read_info_ptr, &width, &height, &bits, &mode, NULL, NULL, NULL
  );
  if(bits != 8) {
    Exception::showError("PNG bit depth not supported","PNG Error");
    Exception::throws("PNG");
  }
  DRPngImage* pngImage;
  int components;
  int paletteEntries;
  png_color_struct* palette = NULL;
  int transpEntries;
  png_bytep transp = NULL;
  switch(mode) {
    case PNG_COLOR_TYPE_PALETTE: {
      png_get_PLTE(read_ptr,read_info_ptr,&palette,&paletteEntries);
      png_get_tRNS(read_ptr,read_info_ptr,&transp,&transpEntries,NULL);
      if(transpEntries != 0) {
        pngImage = new DRPngImage(true);
        components = 4;
      } else {
        pngImage = new DRPngImage(false);
        components = 3;
      }
      break;
    }
    case PNG_COLOR_TYPE_RGB: {
      pngImage = new DRPngImage(false);
      components = 3;
      break;
    }
    case PNG_COLOR_TYPE_GRAY: {
      pngImage = new DRPngImage(false,true);
      components = 1;
      break;
    }
    case PNG_COLOR_TYPE_GRAY_ALPHA:
    case PNG_COLOR_TYPE_RGB_ALPHA: {
      pngImage = new DRPngImage(true);
      components = 4;
      break;
    }
    default:
      Exception::showError("PNG format not supported","PNG Error");
      Exception::throws("PNG");
  }
  pngImage->setWidth(width);
  pngImage->setHeight(width);
  int row_stride = width*components;
  png_bytep* pngBuffer = new png_bytep[height];
  unsigned char* buffer = new unsigned char[row_stride*height];
  for(unsigned int ct = 0; ct < height; ct++)
    pngBuffer[ct] = &buffer[ct*row_stride];
  png_read_image(read_ptr,pngBuffer);
  png_read_end(read_ptr,NULL);
  switch(mode) {
    case PNG_COLOR_TYPE_PALETTE: {
      for(unsigned int ct = 0; ct < height; ct++) {
        int ct1 = width-1;
        for(int ct3 = ct1*components; ct3 >= 0; ct3 -= components, ct1--) {
      	  unsigned char idx = pngBuffer[ct][ct1];
          pngBuffer[ct][ct3  ] = palette[idx].red;
          pngBuffer[ct][ct3+1] = palette[idx].green;
          pngBuffer[ct][ct3+2] = palette[idx].blue;
          if(transp && idx < transpEntries)
            pngBuffer[ct][ct3+3] = transp[idx];
          else
            pngBuffer[ct][ct3+3] = 0xff;
        }
      }
      break;
    }
    case PNG_COLOR_TYPE_GRAY: {
      /*
      for(unsigned int ct = 0; ct < height; ct++) {
        for(int ct1 = width-1; ct1 >= 0; ct1--) {
      	  pngBuffer[ct][ct1] = pngBuffer[ct][ct1];
      	}
      }
      */
      break;
    }
    case PNG_COLOR_TYPE_GRAY_ALPHA: {
      for(unsigned int ct = 0; ct < height; ct++) {
        int ct2 = width+width-2;
        for(int ct4 = (width-1)*4; ct4 >= 0; ct4 -= 4, ct2 -=2) {
      	  unsigned char index = pngBuffer[ct][ct2  ];
          unsigned char alpha = pngBuffer[ct][ct2+1];
          pngBuffer[ct][ct4  ] = index;
          pngBuffer[ct][ct4+1] = index;
          pngBuffer[ct][ct4+2] = index;
          pngBuffer[ct][ct4+3] = alpha;
        }
      }
      break;
    }
    case PNG_COLOR_TYPE_RGB_ALPHA: {
      /*
      for(unsigned int ct = 0; ct < height; ct++) {
        for(int ct4 = (width-1)*4; ct4 >= 0; ct4 -= 4) {
          unsigned char r = pngBuffer[ct][ct4  ];
      	  unsigned char g = pngBuffer[ct][ct4+1];
          unsigned char b = pngBuffer[ct][ct4+2];
          unsigned char a = pngBuffer[ct][ct4+3];
          pngBuffer[ct][ct4  ] = r;
          pngBuffer[ct][ct4+1] = g;
          pngBuffer[ct][ct4+2] = b;
          pngBuffer[ct][ct4+3] = a;
      	}
      }
      */
      break;
    }
    case PNG_COLOR_TYPE_RGB: {
      /*
      for(unsigned int ct = 0; ct < height; ct++) {
        for(int ct3 = (width-1)*3; ct3 >= 0; ct3 -= 3) {
          unsigned char r = pngBuffer[ct][ct3  ];
          unsigned char g = pngBuffer[ct][ct3+1];
          unsigned char b = pngBuffer[ct][ct3+2];
          pngBuffer[ct][ct3  ] = r;
          pngBuffer[ct][ct3+1] = g;
          pngBuffer[ct][ct3+2] = b;
      	}
      }
      */
      break;
    }
    default:
      break;
  }
  png_destroy_read_struct(&read_ptr,&read_info_ptr,&end_info_ptr);
  delete pngBuffer;
  pngImage->setBuffer(buffer);
  return pngImage;
}

//
// DR3DSReader
//
DR3DSReader::DR3DSReader(DRZipFile* zip):
  zipFile(zip),
  readFile(INVALID_HANDLE_VALUE),
  chunkPointer(0), chunkEndPointer(0)
{
  bufferPointer = bufferStart = zip->getZipBuffer();
  bufferLength = zip->getZipBufferLength();
  readDataPtr = &readDataFromZip;
  seekDataPtr = &seekDataFromZip;
  dataLengthPtr = &dataLengthFromZip;
  readImagePtr = & readImageFromZip;
}

DR3DSReader::DR3DSReader(HANDLE fileHandle):
  zipFile(NULL), bufferPointer(NULL), bufferStart(NULL), bufferLength(0),
  readFile(fileHandle),
  chunkPointer(0), chunkEndPointer(0)
{
  readDataPtr = &readDataFromFile;
  seekDataPtr = &seekDataFromFile;
  dataLengthPtr = &dataLengthFromFile;
  readImagePtr = &readImageFromFile;
}

void DR3DSReader::readDataFromFile(
  DR3DSReader* reader, LPVOID buffer, DWORD bytesToBeRead, LPDWORD bytesRead
) {
  ReadFile(reader->readFile,buffer,bytesToBeRead,bytesRead,NULL);
}

void DR3DSReader::readDataFromZip(
  DR3DSReader* reader, LPVOID buffer, DWORD bytesToBeRead, LPDWORD bytesRead
) {
  unsigned char* byteBuffer = (unsigned char*)buffer;
  for(unsigned int ct = 0; ct < bytesToBeRead; ct++) {
    byteBuffer[ct] = *(reader->bufferPointer);
    (reader->bufferPointer)++;
  }
  *bytesRead = bytesToBeRead;
}

void DR3DSReader::seekDataFromFile(DR3DSReader* reader, DWORD offset) {
  SetFilePointer(reader->readFile,offset,NULL,FILE_BEGIN);
}

void DR3DSReader::seekDataFromZip(DR3DSReader* reader, DWORD offset) {
  reader->bufferPointer = reader->bufferStart+offset;
}

int DR3DSReader::dataLengthFromFile(DR3DSReader* reader) {
  return GetFileSize(reader->readFile,NULL);
}

int DR3DSReader::dataLengthFromZip(DR3DSReader* reader) {
  return reader->bufferLength;
}

DRImage* DR3DSReader::readImageFromFile(DR3DSReader* reader, char* imageName) {
  reader;
  DRImage* image = NULL;
  if(strstr(imageName,".jpg")) {
    DRJpegFile file(imageName);
    image = file.getImage();
  } else if(strstr(imageName,".png")) {
    DRPngFile file(imageName);
    image = file.getImage();
  }
  return image;
}

DRImage* DR3DSReader::readImageFromZip(DR3DSReader* reader, char* imageName) {
  return reader->zipFile->getImage(imageName);
}

unsigned short DR3DSReader::getChunkID() {
  unsigned short id;
  int chunkLength;
  unsigned long size;
  readData(&id,2,&size);
  readData(&chunkLength,4,&size);
  chunkEndPointer = chunkPointer+chunkLength;
  chunkPointer += 6;
  return id;
}

void DR3DSReader::skip() {
  seekData(chunkPointer = chunkEndPointer);
}

void DR3DSReader::restore() {
  seekData(chunkPointer = 6);
}

unsigned short DR3DSReader::readUShort() {
  unsigned short uShort;
  unsigned long size;
  readData(&uShort,2,&size);
  chunkPointer += 2;
  return uShort;
}

unsigned char DR3DSReader::readUChar() {
  unsigned char uChar;
  unsigned long size;
  readData(&uChar,1,&size);
  chunkPointer++;
  return uChar;
}

void DR3DSReader::read(char* string) {
  unsigned long size;
  int ct = 0;
  do {
    readData(&string[ct],1,&size);
    chunkPointer++;
  } while(string[ct++] != '\0');
}

void DR3DSReader::read(float* array, int arrayLength) {
  unsigned long size;
  arrayLength *= 4;
  readData(array,arrayLength,&size);
  chunkPointer += arrayLength;
}

GLObjects* DR3DSReader::readMeshes(bool bumped) {
  const int INC_OBJECTS_COUNT = 8;
  const int MAX_NAME_LENGTH = 64;
  DSArray<char> materialNames(INC_OBJECTS_COUNT);
  GLObjects* meshes = new GLObjects();
  int fileSize = dataLength();
  if(fileSize == 0xFFFFFFFF) return NULL;
  if(getChunkID() != MAIN3DS) return NULL;
  while(isMoreDataAvailable(fileSize)) {
    switch(getChunkID()) {
      case EDIT3DS: {
        int editEnd = getChunkEndPointer();
        while(isMoreDataAvailable(editEnd)) {
          switch(getChunkID()) {
            case OBJECT3DS: {
              int objectEnd = getChunkEndPointer();
              char objectName[MAX_NAME_LENGTH];
              read(objectName);
              int verticesCount = 0;
              float* vertices = NULL;
              float* mappings = NULL;
              int facesCount = 0;
              unsigned short* faces = NULL;
              while(isMoreDataAvailable(objectEnd)) {
                switch(getChunkID()) {
                  case TRIMESH: {
                    break;
                  }
                  case VERTEXLIST: {
                    verticesCount = readUShort();
                    int arrayLength = verticesCount*3;
                    vertices = new float[arrayLength];
                    read(vertices,arrayLength);
                    break;
                  }
                  case MAPPINGLIST: {
                    int mappingsCount = readUShort();
                    int arrayLength = mappingsCount+mappingsCount;
                    mappings = new float[arrayLength];
                    read(mappings,arrayLength);
                    break;
                  }
                  case FACELIST: {
                    facesCount = readUShort();
                    int arrayLength = facesCount*3;
                    faces = new unsigned short[arrayLength];
                    for(int ct = 0; ct < arrayLength; ct+=3) {
                      faces[ct  ] = readUShort();
                      faces[ct+1] = readUShort();
                      faces[ct+2] = readUShort();
                      readUShort();
                    }
                    break;
                  }
                  case MATERIALLIST: {
                    char* matName = new char[MAX_NAME_LENGTH];
                    read(matName);
                    materialNames.addElement(matName);
                    skip();
                    break;
                  }
                  default: {
                    skip();
                    break;
                  }
                }
              }
              GLShape* shape = new GLShape(
                verticesCount, vertices, NULL, mappings, facesCount, faces
              );
              meshes->addObject(
                bumped? new GLBumpedMesh(shape): new GLMesh(shape)
              );
              if(vertices) delete vertices;
              if(mappings) delete mappings;
              if(faces) delete faces;
              break;
            }
            default: {
              skip();
              break;
            }
          }
        }
        break;
      }
      default: {
        skip();
        break;
      }
    }
  }
  restore();
  while(isMoreDataAvailable(fileSize)) {
    switch(getChunkID()) {
      case EDIT3DS: {
        int editEnd = getChunkEndPointer();
        while(isMoreDataAvailable(editEnd)) {
          switch(getChunkID()) {
            case MATERIALS: {
              GLObject* currentObject = NULL;
              int materialEnd = getChunkEndPointer();
              while(isMoreDataAvailable(materialEnd)) {
                switch(getChunkID()) {
                  case MATERIALNAME: {
                    char materialName[MAX_NAME_LENGTH];
                    read(materialName);
                    currentObject = meshes->getFirstObject();
                    int matCounter = 0;
                    while(
                      currentObject &&
                      strcmp(
                        materialName,materialNames.getElement(matCounter)
                      ) != 0
                    ) {
                      currentObject = meshes->getNextObject();
                      matCounter++;
                    }
                    break;
                  }
                  case MATAMBIENT: {
                    if(currentObject) {
                      getChunkID();
                      float r = readUChar()/255.0f;
                      float g = readUChar()/255.0f;
                      float b = readUChar()/255.0f;
                      dynamic_cast<GLMesh*>(
                        currentObject
                      )->getMaterial().setAmbient(r,g,b);
                    } else {
                      skip();
                    }
                    break;
                  }
                  case MATDIFFUSE: {
                    if(currentObject) {
                      getChunkID();
                      float r = readUChar()/255.0f;
                      float g = readUChar()/255.0f;
                      float b = readUChar()/255.0f;
                      dynamic_cast<GLMesh*>(
                        currentObject
                      )->getMaterial().setDiffuse(r,g,b);
                    } else {
                      skip();
                    }
                    break;
                  }
                  case MATSPECULAR: {
                    if(currentObject) {
                      getChunkID();
                      float r = readUChar()/255.0f;
                      float g = readUChar()/255.0f;
                      float b = readUChar()/255.0f;
                      dynamic_cast<GLMesh*>(
                        currentObject
                      )->getMaterial().setSpecular(r,g,b);
                    } else {
                      skip();
                    }
                    break;
                  }
                  case MATTEXTURE: {
                    break;
                  }
                  case MATTEXNAME: {
                    char textureName[MAX_NAME_LENGTH];
                    read(textureName);
                    if(currentObject) {
                      DRImage* image = readImage(textureName);
                      if(image) {
                        GLTexture* texture = new GLTexture(*image);
                        dynamic_cast<GLMesh*>(
                          currentObject
                        )->getMaterial().setDiffuseTexture(texture);
                        delete image;
                      }
                    }
                    break;
                  }
                  default: {
                    skip();
                    break;
                  }
                }
              }
              break;
            }
            default: {
              skip();
              break;
            }
          }
        }
        break;
      }
      default: {
        skip();
        break;
      }
    }
  }
  return meshes;
}

//
// DRResource
//
DRResource::DRResource(int id) throw(Exception) {
  resource = FindResource(NULL,MAKEINTRESOURCE(id),MAKEINTRESOURCE(RT_RCDATA));
  if(resource == NULL) Exception::throws("FindResource");
}

DRResource::DRResource(char* name) throw(Exception) {
  resource = FindResource(NULL,name,MAKEINTRESOURCE(RT_RCDATA));
  if(resource == NULL) Exception::throws("FindResource");
}

LPBYTE DRResource::getPointer() throw(Exception) {
  HGLOBAL memory = LoadResource(NULL,resource);
  if(memory == NULL) Exception::throws("LoadResource");
  LPBYTE pointer = (LPBYTE)LockResource(memory);
  if(pointer == NULL) Exception::throws("LockResource");
  return pointer;
}

DRResource::~DRResource() {
}

//
// DRFile
//
DRFile::DRFile(char* filNam) throw(Exception):
  fileName(filNam), file(INVALID_HANDLE_VALUE)
{}

DRFile::~DRFile() {
  closeFile();
}

bool DRFile::openFile() {
  file = CreateFile(
    fileName,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL
  );
  return file != INVALID_HANDLE_VALUE;
}

void DRFile::closeFile() {
  if(file != INVALID_HANDLE_VALUE) {
    CloseHandle(file);
    file = INVALID_HANDLE_VALUE;
  }
}

//
// DRDataFile
//
DRDataFile::DRDataFile(char* filNam) throw(Exception): DRFile(filNam) {
}

DRData* DRDataFile::getData() {
  if(openFile()) {
    DWORD size = GetFileSize(file,NULL);
    unsigned char* buffer = new unsigned char[size];
    ReadFile(file,buffer,size,&size,NULL);
    closeFile();
    return new DRData(size,buffer);
  }
  return NULL;
}

//
// DRJpegFile
//
DRJpegFile::DRJpegFile(char* filNam) throw(Exception):
  DRFile(filNam), DRJpegReader()
{}

DRJpegImage* DRJpegFile::getImage(int alphaColor, int tolerance) {
  my_source_mgr_fil src;
  src.file = this;
  return DRJpegReader::readImage(
    &cinfo,&src.pub,
    init_source_fil,fill_input_buffer_fil,
    skip_input_data_fil,term_source_fil,
    alphaColor, tolerance
  );
}

//
// DRPngFile
//
DRPngFile::DRPngFile(char* filNam) throw(Exception):
  DRDataFile(filNam), DRPngReader()
{}

DRPngImage* DRPngFile::getImage() {
  DRData* data = getData();
  if(data) {
    setReadPointer(data->getBuffer());
    DRPngImage* img = DRPngReader::readImage();
    delete data;
    return img;
  }
  return NULL;
}

//
// DRZipFile
//
DRZipFile::DRZipFile(char* zipFileName) throw(Exception):
  zipBuffer(NULL), zipBufferLength(0), DRJpegReader()
{
  if(zipFileName) {
    if((fileHandler = unzOpen(zipFileName)) == NULL) {
      char fileName[512];
      strcpy(fileName,zipFileName);
      strcat(fileName,".zip");
      if((fileHandler = unzOpen(fileName)) == NULL)
        Exception::throws("ZIP file not found.");
    }
  }
}

DRZipFile::~DRZipFile() {
  unzClose(fileHandler);
}

bool DRZipFile::createZipBuffer(char* fileName) {
  if(findZippedFile(fileName) && openZippedFile()) {
    zipBufferLength = getZippedFileSize();
    zipBuffer = new unsigned char[zipBufferLength];
    readZippedFile();
    return true;
  }
  return false;
}

void DRZipFile::deleteZipBuffer(bool deleteBuffer) {
  if(zipBuffer) {
    if(deleteBuffer) delete zipBuffer;
    zipBuffer = NULL;
    zipBufferLength = 0;
  }
  closeZippedFile();
}

DRData* DRZipFile::getData(char* fileName) {
  if(createZipBuffer(fileName)) {
    DRData* data = new DRData(zipBufferLength,zipBuffer);
    deleteZipBuffer(false);
    return data;
  }
  return NULL;
}

DRImage* DRZipFile::getImage(char* fileName) {
  if(strstr(fileName,".jpg")) {
    return getJpegImage(fileName);
  } else if(strstr(fileName,".png")) {
    return getPngImage(fileName);
  }
  return NULL;
}

DRJpegImage* DRZipFile::getJpegImage(char* fileName, int alphaColor, int tol) {
  my_source_mgr_zip src;
  src.zipFile = this;
  src.zippedFileName = fileName;
  return DRJpegReader::readImage(
    &cinfo,&src.pub,
    init_source_zip,fill_input_buffer_zip,
    skip_input_data_zip,term_source_zip,
    alphaColor,tol
  );
}

DRPngImage* DRZipFile::getPngImage(char* fileName) {
  if(createZipBuffer(fileName)) {
    setReadPointer(zipBuffer);
    DRPngImage* pi = DRPngReader::readImage();
    deleteZipBuffer();
    return pi;
  }
  return NULL;
}

GLObjects* DRZipFile::getMeshes(char* fileName, bool bumped) {
  if(createZipBuffer(fileName)) {
    unsigned char* modelZipBuffer = zipBuffer;
    DR3DSReader reader(this);
    GLObjects* meshes = reader.readMeshes(bumped);
    if(modelZipBuffer) delete modelZipBuffer;
    return meshes;
  }
  return NULL;
}

FMSample* DRZipFile::getSoundSample(char* fileName) {
  if(createZipBuffer(fileName)) {
    FMSample* ret = new FMSample(zipBufferLength,zipBuffer);
    deleteZipBuffer();
    return ret;
  }
  return NULL;
}

FM3DSample* DRZipFile::get3DSoundSample(char* fileName) {
  if(createZipBuffer(fileName)) {
    FM3DSample* ret = new FM3DSample(zipBufferLength,zipBuffer);
    deleteZipBuffer();
    return ret;
  }
  return NULL;
}

FMMusic* DRZipFile::getMusic(char* fileName) {
  if(createZipBuffer(fileName)) {
    FMMusic* ret = new FMMusic(zipBufferLength,zipBuffer);
    deleteZipBuffer();
    return ret;
  }
  return NULL;
}

//
// DRPngResource
//
DRPngResource::DRPngResource(int id) throw(Exception): DRResource(id) {
}

DRPngImage* DRPngResource::getImage() throw(Exception) {
  setReadPointer(getPointer());
  return DRPngReader::readImage();
}

//
// DRJpegResource
//
DRJpegResource::DRJpegResource(int id) throw(Exception):
  DRResource(id), DRJpegReader()
{}

DRJpegImage* DRJpegResource::getImage(int alphaColor, int tolerance) {
  my_source_mgr_res src;
  src.resource = this;
  return DRJpegReader::readImage(
    &cinfo,&src.pub,
    init_source_res,fill_input_buffer_res,
    skip_input_data_res,term_source_res,
    alphaColor,tolerance
  );
}

//
// DRSoundResource
//
DRSoundResource::DRSoundResource(int id) throw(Exception): DRResource(id) {
}

FMSample* DRSoundResource::getSoundSample() {
  return new FMSample(getSize(),getPointer());
}

FM3DSample* DRSoundResource::get3DSoundSample() {
  return new FM3DSample(getSize(),getPointer());
}

//
// DRMusicResource
//
DRMusicResource::DRMusicResource(int id) throw(Exception): DRResource(id) {
}

FMMusic* DRMusicResource::getMusic() {
  return new FMMusic(getSize(),getPointer());
}

//
// DRData
//
DRData::DRData(int sz, unsigned char* buf, bool ob):
  size(sz), buffer(buf), ownsBuffer(ob)
{}

DRData::~DRData() {
  if(ownsBuffer && buffer) delete buffer;
}

//
// DRImage
//
DRImage::DRImage(bool ha, bool gs, bool ob):
  hasAlpha(ha), grayScale(gs), width(0), height(0),
  ownsBuffer(ob), buffer(NULL)
{}

DRImage::DRImage(int w, int h, unsigned char* b, bool ha, bool gs, bool ob):
  width(w), height(h), hasAlpha(ha), grayScale(gs),
  buffer(b), ownsBuffer(ob)
{}

DRImage::~DRImage() {
  if(ownsBuffer && buffer) delete buffer;
}

void DRImage::addAlpha(DRImage& img) {
  const int rgbaLen = 4;
  if(hasAlpha) {
    if(img.isGrayScale()) {
      int rowLen = width*rgbaLen;
      int grayScaleIndex = 0;
      for(int rowBase = 0; rowBase < rowLen*height; rowBase += rowLen) {
        for(int rowPos = rowBase; rowPos < rowBase+rowLen; rowPos += rgbaLen) {
          int rowPosAlpha = rowPos+3;
          buffer[rowPosAlpha] = img.getBuffer()[grayScaleIndex++];
        }
      }
    } else {
      int rowLen = width*rgbaLen;
      for(int rowBase = 0; rowBase < rowLen*height; rowBase += rowLen) {
        for(int rowPos = rowBase; rowPos < rowBase+rowLen; rowPos += rgbaLen) {
          int rowPosAlpha = rowPos+3;
          buffer[rowPosAlpha] = img.getBuffer()[rowPosAlpha];
        }
      }
    }
  } else {
    unsigned char* newBuffer = new unsigned char[width*height*rgbaLen];
    hasAlpha = true;
    int rowLen = width*rgbaLen;
    int oldRowPos = 0;
    if(img.isGrayScale()) {
      int grayScaleIndex = 0;
      if(isGrayScale()) {
        setGrayScale(false);
        for(int rowBase = 0; rowBase < rowLen*height; rowBase += rowLen) {
          for(int rowPos = rowBase; rowPos < rowBase+rowLen; rowPos += rgbaLen){
            newBuffer[rowPos  ] = buffer[grayScaleIndex];
            newBuffer[rowPos+1] = buffer[grayScaleIndex];
            newBuffer[rowPos+2] = buffer[grayScaleIndex];
            newBuffer[rowPos+3] = img.getBuffer()[grayScaleIndex++];
          }
        }
      } else {
        for(int rowBase = 0; rowBase < rowLen*height; rowBase += rowLen) {
          for(int rowPos = rowBase; rowPos < rowBase+rowLen; rowPos += rgbaLen){
            newBuffer[rowPos  ] = buffer[oldRowPos++];
            newBuffer[rowPos+1] = buffer[oldRowPos++];
            newBuffer[rowPos+2] = buffer[oldRowPos++];
            newBuffer[rowPos+3] = img.getBuffer()[grayScaleIndex++];
          }
        }
      }
    } else {
      if(isGrayScale()) {
        setGrayScale(false);
        int grayScaleIndex = 0;
        for(int rowBase = 0; rowBase < rowLen*height; rowBase += rowLen) {
          for(int rowPos = rowBase; rowPos < rowBase+rowLen; rowPos += rgbaLen){
            newBuffer[rowPos  ] = buffer[grayScaleIndex];
            newBuffer[rowPos+1] = buffer[grayScaleIndex];
            newBuffer[rowPos+2] = buffer[grayScaleIndex++];
            newBuffer[rowPos+3] = img.getBuffer()[rowPos+3];
          }
        }
      } else {
        for(int rowBase = 0; rowBase < rowLen*height; rowBase += rowLen) {
          for(int rowPos = rowBase; rowPos < rowBase+rowLen; rowPos += rgbaLen){
            newBuffer[rowPos  ] = buffer[oldRowPos++];
            newBuffer[rowPos+1] = buffer[oldRowPos++];
            newBuffer[rowPos+2] = buffer[oldRowPos++];
            newBuffer[rowPos+3] = img.getBuffer()[rowPos+3];
          }
        }
      }
    }
    if(ownsBuffer)
      delete buffer;
    else
      ownsBuffer = true;
    buffer = newBuffer;
  }
}

void DRImage::convertGray2RGBA() {
  if(grayScale) {
    const int rgbaLen = 4;
    hasAlpha = true;
    grayScale = false;
    unsigned char* newBuffer = new unsigned char[width*height*rgbaLen];
    int rowLen = width*rgbaLen;
    int grayScaleIndex = 0;
    for(int rowBase = 0; rowBase < rowLen*height; rowBase += rowLen) {
      for(int rowPos = rowBase; rowPos < rowBase+rowLen; rowPos += rgbaLen) {
        unsigned char grayShade = buffer[grayScaleIndex++];
        newBuffer[rowPos  ] = grayShade;
        newBuffer[rowPos+1] = grayShade;
        newBuffer[rowPos+2] = grayShade;
        newBuffer[rowPos+3] = grayShade;
      }
    }
    if(ownsBuffer)
      delete buffer;
    else
      ownsBuffer = true;
    buffer = newBuffer;
  }
}

void DRImage::convertGray2111A() {
  if(grayScale) {
    const int rgbaLen = 4;
    hasAlpha = true;
    grayScale = false;
    unsigned char* newBuffer = new unsigned char[width*height*rgbaLen];
    int rowLen = width*rgbaLen;
    int grayScaleIndex = 0;
    for(int rowBase = 0; rowBase < rowLen*height; rowBase += rowLen) {
      for(int rowPos = rowBase; rowPos < rowBase+rowLen; rowPos += rgbaLen) {
        unsigned char grayShade = buffer[grayScaleIndex++];
        newBuffer[rowPos  ] = 255;
        newBuffer[rowPos+1] = 255;
        newBuffer[rowPos+2] = 255;
        newBuffer[rowPos+3] = grayShade;
      }
    }
    if(ownsBuffer)
      delete buffer;
    else
      ownsBuffer = true;
    buffer = newBuffer;
  }
}

//
// DRJpegImage
//
DRJpegImage::DRJpegImage(int ac, bool ob): DRImage(ac >= 0, false, ob) {
}

//
// DRPngImage
//
DRPngImage::DRPngImage(bool ha, bool gs, bool ob): DRImage(ha,gs,ob) {
}

//
// DR3DSFile
//
DR3DSFile::DR3DSFile(char* filNam): DRFile(filNam) {
}

GLObjects* DR3DSFile::getMeshes(bool bumped) {
  if(openFile()) {
    DR3DSReader reader(file);
    GLObjects* meshes = reader.readMeshes(bumped);
    closeFile();
    return meshes;
  }
  return NULL;
}
