/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

//
// DSStaticArray
//
template <class T> DSStaticArray<T>::DSStaticArray():
	increment(1), capacity(0), size(0), elements(NULL)
{}

template <class T> DSStaticArray<T>::DSStaticArray(int cap):
	increment(cap), capacity(cap), size(0), elements(new T[cap])
{}

template <class T> DSStaticArray<T>::~DSStaticArray() {
 	if(elements) delete elements;
}

template <class T> void DSStaticArray<T>::incrementCapacity() {
  int newCapacity = capacity + increment;
  T* slot = new T[newCapacity];
  for(int ct = 0; ct < capacity; ct++) slot[ct] = elements[ct];
  capacity = newCapacity;
  if(elements) delete elements;
  elements = slot;
}

//
// DSArray
//
template <class T, bool ownsElements> DSArray<T,ownsElements>::DSArray():
  DSStaticArray<T*>()
{}

template <class T, bool ownsElements> DSArray<T,ownsElements>::DSArray(int cap):
  DSStaticArray<T*>(cap)
{
	for(int ct = 0; ct < cap; ct++) elements[ct] = NULL;
}

template <class T, bool ownsElements> DSArray<T,ownsElements>::~DSArray() {
  if(ownsElements)
    for(int ct = 0; ct < size; ct++) if(elements[ct]) delete elements[ct];
}

template <class T, bool ownsElements>
void DSArray<T,ownsElements>::incrementCapacity() {
  int newCapacity = capacity + increment;
  T** slot = new T*[newCapacity];
  for(int ct = 0; ct < capacity; ct++) slot[ct] = elements[ct];
  for(int ct = capacity; ct < newCapacity; ct++) slot[ct] = NULL;
  capacity = newCapacity;
  if(elements) delete elements;
  elements = slot;
}

template <class T, bool ownsElements>
void DSArray<T,ownsElements>::removeElement(int idx) {
  if(idx >= 0 && idx < size) {
    if(ownsElements) if(elements[idx]) delete elements[idx];
    size--;
    for(int ct = idx; ct < size; ct++) elements[ct] = elements[ct+1];
    elements[size] = NULL;
  }
}

template <class T, bool ownsElements>
bool DSArray<T,ownsElements>::removeElement(T* element) {
  bool removed = false;
  for(int idx = 0; idx < size; idx++) {
    if(elements[idx] == element) {
      if(ownsElements) if(elements[idx]) delete elements[idx];
      size--;
      for(int ct = idx; ct < size; ct++) elements[ct] = elements[ct+1];
      elements[size] = NULL;
      removed = true;
      break;
    }
  }
  return removed;
}

template <class T, bool ownsElements>
void DSArray<T,ownsElements>::deleteElements() {
  if(ownsElements)
    for(int ct = 0; ct < size; ct++) if(elements[ct]) delete elements[ct];
  size = 0;
}

//
// DSSortableObject
//

//
// DSSortedArray
//
template <class T, bool ownsElements = true>
DSSortedArray<T,ownsElements>::DSSortedArray(): DSArray<T,ownsElements>() {
}

template <class T, bool ownsElements = true>
DSSortedArray<T,ownsElements>::DSSortedArray(int cap):
  DSArray<T,ownsElements>(cap)
{}

template <class T, bool ownsElements = true>
void DSSortedArray<T,ownsElements>::quickSort(
  int low,int high
) {
  if(low < high) {
    int lo = low;
    int hi = high+1;
    while(true) {
      while(
        (getElement(++lo)->getSortIndex() >= getElement(low)->getSortIndex()) &&
        (lo < high)
      );
      while(
        (getElement(--hi)->getSortIndex() <= getElement(low)->getSortIndex()) &&
        (hi > low)
      );
      if(lo < hi)
        setElement(hi,setElement(lo,getElement(hi)));
      else
        break;
    }
    setElement(low,setElement(hi,getElement(low)));
    quickSort(low,hi-1);
    quickSort(hi+1,high);
  }
}

//
// DSChainNode
//
DSChainNode::DSChainNode(): next(0) {
}

DSChainNode::~DSChainNode() {
}

//
// DSTwoWayChainNode
//
DSTwoWayChainNode::DSTwoWayChainNode(): prev(0), DSChainNode() {
}

//
// DSChain
//
DSChain::DSChain():
  length(0), first(0), last(0), pointer(0)
{}

DSChain::~DSChain() {
  deleteAll();
}

DSChainNode* DSChain::nextNode() {
  return pointer? pointer = pointer->getNextNode(): 0;
}

void DSChain::addNodeFirst(DSChainNode* node) {
  if(first) {
    node->setNextNode(first);
  } else {
    last = pointer = node;
  }
  first = node;
  length++;
}

void DSChain::addNodeLast(DSChainNode* node) {
  if(first) {
    last->setNextNode(node);
  } else {
    first = pointer = node;
  }
  last = node;
  length++;
}

void DSChain::insertNodeBehind(DSChainNode* node) {
  if(pointer) {
    DSChainNode* nextNode = pointer->getNextNode();
    pointer->setNextNode(node);
    node->setNextNode(nextNode);
  } else {
    node->setNextNode(first);
    first = node;
  }
  if(last == pointer) last = node;
  length++;
}

bool DSChain::deleteNode(DSChainNode* node) {
  DSChainNode* prevNode = 0;
  DSChainNode* thisNode = first;
  while(thisNode) {
    if(thisNode == node) {
      pointer = thisNode->getNextNode();
      if(!pointer) last = prevNode;
    	if(prevNode) {
      	prevNode->setNextNode(pointer);
      } else {
      	first = thisNode->getNextNode();
      }
      length--;
      delete node;
      return true;
    }
    prevNode = thisNode;
    thisNode = thisNode->getNextNode();
  }
  return false;
}

void DSChain::deleteAll() {
  while((pointer = first) != 0) {
    first = pointer->getNextNode();
    delete pointer;
  }
  last = 0;
  length = 0;
}

//
// DSTwoWayChain
//
DSTwoWayChainNode* DSTwoWayChain::prevNode() {
	return pointer?
  	dynamic_cast<DSTwoWayChainNode*>(
    	pointer = dynamic_cast<DSTwoWayChainNode*>(pointer)->getPrevNode()
    ): 0;
}

void DSTwoWayChain::addNodeFirst(DSTwoWayChainNode* node) {
	DSTwoWayChainNode* next = dynamic_cast<DSTwoWayChainNode*>(first);
	DSChain::addNodeFirst(node);
  dynamic_cast<DSTwoWayChainNode*>(first)->setPrevNode(next);
}

void DSTwoWayChain::addNodeLast(DSTwoWayChainNode* node) {
	DSTwoWayChainNode* prev = dynamic_cast<DSTwoWayChainNode*>(last);
	DSChain::addNodeLast(node);
  dynamic_cast<DSTwoWayChainNode*>(last)->setPrevNode(prev);
}

void DSTwoWayChain::insertNodeBefore(DSTwoWayChainNode* node) {
	if(pointer) {
  	DSTwoWayChainNode* prevNode =
    	dynamic_cast<DSTwoWayChainNode*>(pointer)->getPrevNode()
    ;
    if(prevNode) prevNode->setNextNode(node);
    node->setPrevNode(prevNode);
    node->setNextNode(pointer);
    dynamic_cast<DSTwoWayChainNode*>(pointer)->setPrevNode(node);
  } else {
  	node->setPrevNode(dynamic_cast<DSTwoWayChainNode*>(last));
    if(last) last->setNextNode(node);
    last = node;
  }
  if(first == pointer) first = node;
  length++;
}

void DSTwoWayChain::insertNodeBehind(DSTwoWayChainNode* node) {
	DSTwoWayChainNode* prev = dynamic_cast<DSTwoWayChainNode*>(pointer);
	DSTwoWayChainNode* next =
  	dynamic_cast<DSTwoWayChainNode*>(prev->getNextNode())
  ;
	DSChain::insertNodeBehind(node);
  node->setPrevNode(prev);
  if(next) next->setPrevNode(node);
}

bool DSTwoWayChain::deleteNode(DSTwoWayChainNode* node) {
	DSTwoWayChainNode* prev = node->getPrevNode();
	DSTwoWayChainNode* next = node->getNextNode();
	bool r = DSChain::deleteNode(node);
  if(r && next) next->setPrevNode(prev);
  return r;
}

//
// DSListNode
//
template <class T, bool ownsElements>
DSListNode<T,ownsElements>::DSListNode(T* e): element(e), DSChainNode() {
}

template <class T, bool ownsElements>
DSListNode<T,ownsElements>::~DSListNode() {
  if(ownsElements && element) delete element;
}

//
// DSTwoWayListNode
//
template <class T, bool ownsElements>
DSTwoWayListNode<T,ownsElements>::DSTwoWayListNode(T* e):
	DSListNode<T,ownsElements>(e)
{}

//
// DSList
//
template <class T, bool ownsElements> T* DSList<T,ownsElements>::thisElement() {
  DSListNode<T,ownsElements>* node =
    dynamic_cast<DSListNode<T,ownsElements>*>(getThisNode())
  ;
  return node? node->getElement(): 0;
}

template <class T, bool ownsElements> T* DSList<T,ownsElements>::nextElement() {
  DSListNode<T,ownsElements>* node =
    dynamic_cast<DSListNode<T,ownsElements>*>(nextNode())
  ;
  return node? node->getElement(): 0;
}

template <class T, bool ownsElements>
DSListNode<T,ownsElements>* DSList<T,ownsElements>::addElement(T* e) {
  DSListNode<T,ownsElements>* node = new DSListNode<T,ownsElements>(e);
  addNodeLast((DSChainNode*)node);
  return node;
}

template <class T, bool ownsElements>
DSListNode<T,ownsElements>* DSList<T,ownsElements>::insertElement(T* e) {
  DSListNode<T,ownsElements>* node = new DSListNode<T,ownsElements>(e);
  insertNodeBehind(node);
  return node;
}

template <class T, bool ownsElements>
bool DSList<T,ownsElements>::deleteElement(T* e) {
  for(
    DSListNode<T,ownsElements>* node = reset();
    node != 0;
    node = nextNode()
  )
    if(node->getElement() == e) return deleteNode(node)
  ;
  return false;
}

template <class T, bool ownsElements>
bool DSList<T,ownsElements>::removeElement(T* e) {
  for(
    DSListNode<T,ownsElements>* node = reset();
    node != 0;
    node = nextNode()
  )
    if(node->getElement() == e) return removeNode(node)
  ;
  return false;
}

template <class T, bool ownsElements>
bool DSList<T,ownsElements>::removeNode(DSListNode<T,ownsElements>* n) {
  n->setElement(0);
  return DSChain::deleteNode(n);
}

//
// DSTwoWayList
//
template <class T, bool ownsElements>
T* DSTwoWayList<T,ownsElements>::prevElement() {
	DSTwoWayListNode<T,ownsElements>* node =
  	(DSTwoWayListNode<T,ownsElements>*)prevNode()
  ;
  return node? node->getElement(): NULL;
}

template <class T, bool ownsElements>
DSTwoWayListNode<T,ownsElements>* DSTwoWayList<T,ownsElements>::addElement(
	T* e
) {
	DSTwoWayListNode<T,ownsElements>* node =
  	new DSTwoWayListNode<T,ownsElements>(e)
  ;
  addNodeFirst(node);
  return node;
}

template <class T, bool ownsElements>
DSTwoWayListNode<T,ownsElements>* DSTwoWayList<T,ownsElements>::insertElement(
	T* e
) {
	DSTwoWayListNode<T,ownsElements>* node =
  	new DSTwoWayListNode<T,ownsElements>(e)
  ;
  insertNodeBefore(node);
  return node;
}

template <class T, bool ownsElements>
bool DSTwoWayList<T,ownsElements>::swapElements() {
  DSTwoWayListNode<T,ownsElements>* thisNode =
    dynamic_cast<DSTwoWayListNode<T,ownsElements>*>(getThisNode())
  ;
  if(thisNode == NULL) return false;
  DSListNode<T,ownsElements>* prevNode = thisNode->getPrevNode();
  if(prevNode == NULL) return false;
  T* element = thisNode->getElement();
  thisNode->setElement(prevNode->getElement());
  prevNode->setElement(element);
  return true;
}

template <class T, bool ownsElements>
bool DSTwoWayList<T,ownsElements>::removeNode(
	DSTwoWayListNode<T,ownsElements>* n
) {
	n->setElement(NULL);
	return DSTwoWayChain::deleteNode(n);
}

