/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "small/sc.h"

#include "glconsole.h"

#include "datasets.cpp"

//
// GLConsole
//
GLConsole::GLConsole(): currentRow(0), currentCol(0), interpreters(2) {
  for(int ct = 0; ct < CONSOLE_ROWS; ct++) screen[ct][0] = '\0';
  println("* SimulAxion Console (c) 2002 Leonardo Boselli    *");
  println("*  \"Small\"  Language (c) 1997-2001 ITB CompuPhase *");
  println("\nWrite 'h' for help.\n");
  printPrompt();
}

GLConsole* GLConsole::console = new GLConsole();

int GLConsole::addInterpreter() {
  int size = interpreters.getSize();
  interpreters.addElement(new SLInterpreter());
  return size;
}

void GLConsole::suspend() {
  if(!scriptInterpreter.isRunning()) {
    int ct = 0;
    for(; ct < interpreters.getSize(); ct++)
      if(interpreters.getElement(ct)->isRunning())
        break;
    if(ct == interpreters.getSize())
      return;
  }
  suspended = true;
  newLine();
  printPrompt();
}

void GLConsole::resume() {
  if(!scriptInterpreter.isRunning()) {
    int ct = 0;
    for(; ct < interpreters.getSize(); ct++)
      if(interpreters.getElement(ct)->isRunning())
        break;
    if(ct == interpreters.getSize())
      return;
  }
  suspended = false;
  newLine();
}

void GLConsole::continueExecution(float deltaT) {
  if(scriptInterpreter.isRunning())
    if(scriptInterpreter.execMain(deltaT) != AMX_ERR_SLEEP)
      printPrompt();
  for(int ct = 1; ct < interpreters.getSize(); ct++) {
    SLInterpreter* interpreter = getInterpreter(ct);
    if(interpreter->isRunning())
      interpreter->execMain(deltaT);
  }
}

void GLConsole::print(char ch) {
  switch(ch) {
    case '\n': {
      newLine();
      drawCursor();
      break;
    }
    case '\b': {
      if(currentCol > 0) {
        currentCol--;
        drawCursor();
      }
      break;
    }
    case '\t': {
      print(" ");
      break;
    }
    default: {
      screen[currentRow][currentCol] = ch;
      if(++currentCol >= CONSOLE_COLS) newLine();
      drawCursor();
      break;
    }
  }
}

void GLConsole::print(char* text) {
  int len = strlen(text);
  for(int ct = 0; ct < len; ct++) print(text[ct]);
}

void GLConsole::keyPressed(char key) {
  switch(key) {
    case '\r': {
      char* commandLine = screen[currentRow];
      newLine();
      if(executeCommand(commandLine)) printPrompt();
      break;
    }
    default: {
      print(key);
      break;
    }
  }
}

bool GLConsole::executeCommand(char* textLine) {
  char commandLine[CONSOLE_COLS+1];
  strlwr(strcpy(commandLine,textLine));
  char* token = strtok(commandLine+2," ");
  if(token) {
    switch(token[0]) {
      case 'h': {
        println("List of valid commands:");
        println("h             : shows this list.");
        println("c \"file name\" : compiles a file (\".sma\" default).");
        println("r \"file name\" : executes a file (\".amx\" default).");
        break;
      }
      case 'c': {
        token = strtok(NULL," ");
        if(token) {
          char* args[] = {"sc",token};
          sc_compile(2,args);
        } else {
          println("Missing file name. Usage: c \"file name\"");
        }
        break;
      }
      case 'r': {
        token = strtok(NULL," ");
        if(token) {
          if(!strrchr(token,'.')) {
            char fileName[32];
            strcat(strcpy(fileName,token),".amx");
            token = fileName;
          }
          if(scriptInterpreter.load(token)) {
            suspended = false;
            int ret = scriptInterpreter.initMain();
            return ret != AMX_ERR_SLEEP;
          } else {
            println("Error loading script file.");
          }
        } else {
          println("Missing file name. Usage: r \"file name\"");
        }
        break;
      }
      default: {
        println("Unknown command.");
        break;
      }
    }
  }
  return true;
}

char* GLConsole::getRow(int idx) {
  if((idx < 0) || (idx >= CONSOLE_ROWS)) return NULL;
  if(idx == CONSOLE_ROWS) return screen[currentRow];
  if((idx += currentRow+1) >= CONSOLE_ROWS) idx -= CONSOLE_ROWS;
  return screen[idx];
}

//
// for SLConsole and SLCompiler
//

int sc_vprintf(const char *message, va_list argptr) {
  char string[128];
  int ret = vsprintf(string,message,argptr);
  GLConsole::get().print(string);
  return ret;
}

int sc_printf(const char *message, ...) {
  va_list argptr;
  va_start(argptr,message);
  int ret = sc_vprintf(message,argptr);
  va_end(argptr);
  return ret;
}