/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glfont.h"

//
// GLFont
//

GLFont::GLFont(int h, int w, int s):
  height(h), width(w), spacing(s? s: (w? w: h>>1))
{
  listBase = glGenLists(getListLength());
}

GLFont::~GLFont() {
  glDeleteLists(getListBase(),getListLength());
}

//
// GLTexturedFont
//
GLTexturedFont::GLTexturedFont(int h, int w, int s, GLTexture* txt):
  GLFont(h,w,s), listOffset(32), texture(txt->acquireReference())
{
  const float charSide = 16/256.0f;
	for(int ct = 0; ct < getListLength(); ct++) {
		float charX = (ct%16)*charSide;
		float charY = (ct/16)*charSide;
		glNewList(getListBase()+ct,GL_COMPILE);
		glBegin(GL_QUADS);
		glTexCoord2f(charX,charY+charSide);
		glVertex2i(0,0);
		glTexCoord2f(charX+charSide,charY+charSide);
		glVertex2i(getWidth()-1,0);
		glTexCoord2f(charX+charSide,charY);
		glVertex2i(getWidth()-1,getHeight()-1);
		glTexCoord2f(charX,charY);
		glVertex2i(0,getHeight()-1);
		glEnd();
		glTranslated(getSpacing(),0,0);
		glEndList();
	}
}

//
// GLBitmapFont
//

GLBitmapFont::GLBitmapFont(
  HDC hDC, char* face, int h, int w, int weight, bool italic
):
  GLFont(h,w)
{
  HGDIOBJ oldFont = GetCurrentObject(hDC,OBJ_FONT);
  HFONT font = CreateFont(
    getHeight(), getWidth(), 0, 0, weight, italic, FALSE, FALSE,
    ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
    ANTIALIASED_QUALITY, VARIABLE_PITCH, face
  );
  SelectObject(hDC,font);
  wglUseFontBitmaps(hDC,0,getListLength(),getListBase());
  DeleteObject(font);
  SelectObject(hDC,oldFont);
}

//
// GLOutlineFont
//

GLOutlineFont::GLOutlineFont(
  HDC hDC, char* face, int h, int w, float depth, int weight, bool italic
):
  GLFont(h,w)
{
  HGDIOBJ oldFont = GetCurrentObject(hDC,OBJ_FONT);
  HFONT font = CreateFont(
    getHeight(), getWidth(), 0, 0, weight, italic, FALSE, FALSE,
    ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
    ANTIALIASED_QUALITY, VARIABLE_PITCH, face
  );
  SelectObject(hDC,font);
  wglUseFontOutlines(
    hDC,0,getListLength(),getListBase(),
    0.0f,depth,WGL_FONT_POLYGONS,glyphMetricsFloat
  );
  DeleteObject(font);
  SelectObject(hDC,oldFont);
}

void GLOutlineFont::destroyObject() {
  delete this;
}

//
// GLText
//

GLText::GLText(char* txt, GLOutlineFont* fnt, GLMaterial* mat):
  text(txt), font((GLOutlineFont*)fnt->acquireReference())
{
  if(!mat) mat = new GLMaterial();
  material = mat->acquireReference();
  computeMaxRadius();
}

GLText::~GLText() {
  font->releaseReference();
  material->releaseReference();
}

void GLText::computeMaxRadius() {
  float w = 0;
  for(unsigned int ct = 0; ct < strlen(text); ct++)
    w += font->getCellWidth(text[ct]);
  setMaxRadius(w*0.5f);
}

void GLText::render(GLCamera& camera) {
  glPushMatrix();
  glMultMatrixf(getTransform());
  font->setPosition(-getMaxRadius(),0,0);
  glPushMatrix();
  material->applyColor();
  glDisable(GL_TEXTURE_2D);
  font->initialize();
  font->drawText(strlen(text),text);
  glEnable(GL_TEXTURE_2D);
  glPopMatrix();
  if(material->hasEnvironment()) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE,GL_ONE);
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    material->applyEnvironment();
    font->drawText(strlen(text),text);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    glDisable(GL_BLEND);
  }
  glFrontFace(GL_CCW);
  glPopMatrix();
}
