/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "gllight.h"

#include <stdlib.h> // for random()

//
// GLLight
//

GLLight::GLLight(GLTexture* lghTxt, float lghSiz):
  texture(lghTxt->acquireReference()), lightSize(lghSiz), color(1,1,1),
  isDistant(false)
{
  setSpot();
  setAttenuation();
  setMaxRadius(0);
  setTransparent(true);
}

GLLight::~GLLight() {
  texture->releaseReference();
}

void GLLight::applyPropertiesTo(int index) {
  int light = GL_LIGHT0+index;
  glEnable(light);
  glLighti(light,GL_SPOT_CUTOFF,spot.cutoff);
  glLightf(light,GL_SPOT_EXPONENT,spot.exponent);
  glLighti(light,GL_CONSTANT_ATTENUATION,attenuation.constant);
  glLighti(light,GL_LINEAR_ATTENUATION,attenuation.linear);
  glLighti(light,GL_QUADRATIC_ATTENUATION,attenuation.quadratic);
  GLfloat zero[] = {0,0,0,1};
  glLightfv(light,GL_AMBIENT,zero);
  glLightfv(light,GL_DIFFUSE,color.getFloatArray());
  glLightfv(light,GL_SPECULAR,color.getFloatArray());
  applyPositionTo(index);
}

void GLLight::applyPositionTo(int index) {
  int light = GL_LIGHT0+index;
  glPushMatrix();
  glMultMatrixf(getTransform());
  GLfloat zero[] = {0,0,0,1};
  glLightfv(light,GL_POSITION,zero);
  GLfloat direc[] =
    {direction.getX(),direction.getY(),direction.getZ(), isDistant? 0: 1};
  glLightfv(light,GL_SPOT_DIRECTION,direc);
  glPopMatrix();
}

void GLLight::render(GLCamera& camera) {
  glPushMatrix();
  glMultMatrixf(getTransform());
  float modView[16];
  glGetFloatv(GL_MODELVIEW_MATRIX,modView);
  M3Vector p1(modView[0]*lightSize,modView[4]*lightSize,modView[8]*lightSize);
  M3Vector p2(modView[1]*lightSize,modView[5]*lightSize,modView[9]*lightSize);
  glPushAttrib(GL_FOG_BIT);
  glDisable(GL_FOG);
  glDepthMask(GL_FALSE);
  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE,GL_ONE);
  glColor4fv(color.getFloatArray());
  texture->apply();
  glBegin(GL_QUADS);
  glTexCoord2f(0,1);
  glVertex3f(-p1.getX()-p2.getX(),-p1.getY()-p2.getY(),-p1.getZ()-p2.getZ());
  glTexCoord2f(1,1);
  glVertex3f(+p1.getX()-p2.getX(),+p1.getY()-p2.getY(),+p1.getZ()-p2.getZ());
  glTexCoord2f(1,0);
  glVertex3f(+p1.getX()+p2.getX(),+p1.getY()+p2.getY(),+p1.getZ()+p2.getZ());
  glTexCoord2f(0,0);
  glVertex3f(-p1.getX()+p2.getX(),-p1.getY()+p2.getY(),-p1.getZ()+p2.getZ());
  glEnd();
  glDepthMask(GL_TRUE);
  glDisable(GL_BLEND);
  glEnable(GL_LIGHTING);
  glPopAttrib();
  glPopMatrix();
}

//
// GLSunLight
//

GLSunLight::GLSunLight(
  GLTexture* lghTxt, float lghSiz, M3Vector& direc,
  GLTexture* flareTxt, int flrCount, float flareSz
):
  GLLight(lghTxt,lghSiz*cameraDistance),
  flaresCount(flrCount),flaresSize(flareSz),flares(NULL),
  flaresTexture(flareTxt? flareTxt->acquireReference(): NULL)
{
  setDirectional(direc.normalize());
  if(flaresCount > 0) {
    flares = new GLFlare[flaresCount];
    for(int ct = 0; ct < flaresCount; ct++) {
      flares[ct].position = (ct+ct+2)/(float)(flaresCount);
      if(ct%2)
        flares[ct].size =
          0.5f*(flaresCount-ct)/(float)(flaresCount)+random(64)/256.0;
      else
        flares[ct].size = 0.5f*ct/(float)(flaresCount)+random(256)/256.0;
      flares[ct].color.set(
        1-random(128)/256.0,1-random(128)/256.0,1-random(128)/256.0
      );
    }
  }
}

GLSunLight::~GLSunLight() {
  if(flaresTexture) flaresTexture->releaseReference();
}

void GLSunLight::render(GLCamera& camera) {
  bool showFlares = false;
  if(flaresCount > 0) {
    GLdouble modelMatrix[16];
    glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
    GLdouble projMatrix[16];
    glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT,viewport);
    GLdouble winX, winY, winZ;
    if(
      gluProject(
        getPosition().getX(),getPosition().getY(),getPosition().getZ(),
        modelMatrix,projMatrix,viewport,
        &winX,&winY,&winZ
      ) == GL_TRUE
    ) {
      float depthValue;
      glReadPixels(winX,winY,1,1,GL_DEPTH_COMPONENT,GL_FLOAT,&depthValue);
      showFlares = (depthValue == 1.0f);
    }
  }
  float wrlModView[16];
  glGetFloatv(GL_MODELVIEW_MATRIX,wrlModView);
  glPushMatrix();
  glMultMatrixf(getTransform());
  float modView[16];
  glGetFloatv(GL_MODELVIEW_MATRIX,modView);
  M3Vector p1(modView[0]*lightSize,modView[4]*lightSize,modView[8]*lightSize);
  M3Vector p2(modView[1]*lightSize,modView[5]*lightSize,modView[9]*lightSize);
  glPushAttrib(GL_FOG_BIT);
  glDisable(GL_FOG);
  glDepthMask(GL_FALSE);
  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE,GL_ONE);
  glColor4fv(color.getFloatArray());
  texture->apply();
  glBegin(GL_QUADS);
  glTexCoord2f(0,1);
  glVertex3f(-p1.getX()-p2.getX(),-p1.getY()-p2.getY(),-p1.getZ()-p2.getZ());
  glTexCoord2f(1,1);
  glVertex3f(+p1.getX()-p2.getX(),+p1.getY()-p2.getY(),+p1.getZ()-p2.getZ());
  glTexCoord2f(1,0);
  glVertex3f(+p1.getX()+p2.getX(),+p1.getY()+p2.getY(),+p1.getZ()+p2.getZ());
  glTexCoord2f(0,0);
  glVertex3f(-p1.getX()+p2.getX(),-p1.getY()+p2.getY(),-p1.getZ()+p2.getZ());
  glEnd();
  glDepthMask(GL_TRUE);
  if(showFlares) {
    glDisable(GL_DEPTH_TEST);
    glPushMatrix();
    glLoadMatrixf(wrlModView);
    M3Vector eyePosition = camera.getPosition();
    M3Vector lightDir(getPositionX(),getPositionY(),getPositionZ());
    lightDir.subtract(eyePosition);
    M3Vector lightProjection(lightDir);
    M3Vector viewDir(
      camera.getViewDirectionX(),
      camera.getViewDirectionY(),
      camera.getViewDirectionZ()
    );
    viewDir.normalize();
    lightDir.normalize();
    float luminosity = lightDir.dot(viewDir)*.25;
    float viewDirLength = camera.getClipPlaneNear()+camera.getClipPlaneNear();
    viewDir.scale(viewDirLength);
    lightProjection.scale(
      viewDirLength*viewDirLength/viewDir.dot(lightProjection)
    ).add(eyePosition);
    M3Vector& flareDir = viewDir;
    flareDir.add(eyePosition).subtract(lightProjection);
    flaresTexture->apply();
    glBegin(GL_QUADS);
    for(int ct = 0; ct < flaresCount; ct++) {
      float flareSize = flares[ct].size*flaresSize;
      p1 = M3Vector(
        modView[0]*flareSize,modView[4]*flareSize,modView[8]*flareSize
      );
      p2 = M3Vector(
        modView[1]*flareSize,modView[5]*flareSize,modView[9]*flareSize
      );
      glColor3f(
        flares[ct].color.getRed()*luminosity*color.getRed(),
        flares[ct].color.getGreen()*luminosity*color.getGreen(),
        flares[ct].color.getBlue()*luminosity*color.getBlue()
      );
      M3Vector flarePos(flareDir);
      flarePos.scale(flares[ct].position).add(lightProjection);
      float offU, offV;
      switch(ct%4) {
        default:
        case 0: offU = offV = 0; break;
        case 1: offU = 0.5f; offV = 0; break;
        case 2: offU = 0; offV = 0.5f; break;
        case 3: offU = offV = 0.5f; break;
      };
      glTexCoord2f(offU,offV+0.5f);
      glVertex3f(
        flarePos.getX()-p1.getX()-p2.getX(),
        flarePos.getY()-p1.getY()-p2.getY(),
        flarePos.getZ()-p1.getZ()-p2.getZ()
      );
      glTexCoord2f(offU+0.5f,offV+0.5f);
      glVertex3f(
        flarePos.getX()+p1.getX()-p2.getX(),
        flarePos.getY()+p1.getY()-p2.getY(),
        flarePos.getZ()+p1.getZ()-p2.getZ()
      );
      glTexCoord2f(offU+0.5f,offV);
      glVertex3f(
        flarePos.getX()+p1.getX()+p2.getX(),
        flarePos.getY()+p1.getY()+p2.getY(),
        flarePos.getZ()+p1.getZ()+p2.getZ()
      );
      glTexCoord2f(offU,offV);
      glVertex3f(
        flarePos.getX()-p1.getX()+p2.getX(),
        flarePos.getY()-p1.getY()+p2.getY(),
        flarePos.getZ()-p1.getZ()+p2.getZ()
      );
    }
    glEnd();
    glPopMatrix();
    glEnable(GL_DEPTH_TEST);
  }
  glDisable(GL_BLEND);
  glEnable(GL_LIGHTING);
  glPopAttrib();
  glPopMatrix();
}


