/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glsprite.h"

#include "glwin.h"

#include "datasets.cpp"

//
// GLSpriteRect
//
GLSpriteRect::GLSpriteRect() {
}

GLSpriteRect::GLSpriteRect(float w, float h):
  halfWidth(w*0.5f), halfHeight(h*0.5f)
{}

//
// GLSprite
//
GLSprite::GLSprite() {
}

GLSprite::GLSprite(
  float w, float h, GLMaterial* mat,
  float iLeft, float iBottom, float iRight, float iTop
):
  GLSpriteRect(w,h), left(iLeft), bottom(iBottom), right(iRight), top(iTop)
{
  if(!mat) mat = new GLMaterial();
  material = mat->acquireReference();
  setMaxRadius((w*w+h*h)*0.25f);
}

GLSprite::~GLSprite() {
  material->releaseReference();
}

void GLSprite::render(GLCamera& camera) {
  glPushMatrix();
  glMultMatrixf(getTransform());
  glDisable(GL_CULL_FACE);
  glEnable(GL_ALPHA_TEST);
  if(isTransparent()) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_FALSE);
  }
  material->applyColor();
  material->applyDiffuse();
  glBegin(GL_QUADS);
  glNormal3f(0,0,-1);
  glTexCoord2f(right,top);
  glVertex3f(-getHalfWidth(),-getHalfHeight(),0);
  glTexCoord2f(right,bottom);
  glVertex3f(-getHalfWidth(), getHalfHeight(),0);
  glTexCoord2f(left,bottom);
  glVertex3f( getHalfWidth(), getHalfHeight(),0);
  glTexCoord2f(left,top);
  glVertex3f( getHalfWidth(),-getHalfHeight(),0);
  glEnd();
  glEnable(GL_CULL_FACE);
  if(isTransparent()) {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
  }
  glDisable(GL_ALPHA_TEST);
  glPopMatrix();
}

//
// GLAnimatedSprite
//
GLAnimatedSprite::GLAnimatedSprite(float duration, float sideFrm):
  animated(true), uvStep(1.0f/sideFrm),
  loopDuration(duration), sideFrames(sideFrm)
{}

GLAnimatedSprite::GLAnimatedSprite(
    float w, float h, float duration, float sideFrm, GLMaterial* mat
):
  GLSprite(w,h,mat,0,0,1.0f/sideFrm,1.0f/sideFrm), animated(true),
  uvStep(1.0f/sideFrm), loopDuration(duration), sideFrames(sideFrm)
{}

void GLAnimatedSprite::render(GLCamera& camera) {
  if(isAnimated()) {
    glMatrixMode(GL_TEXTURE);
    glPushMatrix();
    glLoadIdentity();
    double integer;
    int frame = (int)(
      modf(
        GLWin::getElapsedTime()/loopDuration,&integer
      )*sideFrames*sideFrames
    );
    glTranslatef(frame%sideFrames*uvStep,frame/sideFrames*uvStep,0);
    glMatrixMode(GL_MODELVIEW);
  }
  GLSprite::render(camera);
  if(isAnimated()) {
    glMatrixMode(GL_TEXTURE);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
  }
}

//
// GLFace2FaceSprite
//
GLFace2FaceSprite::GLFace2FaceSprite() {
}

GLFace2FaceSprite::GLFace2FaceSprite(
  float w, float h, GLMaterial* mat,
  float iLeft, float iBottom, float iRight, float iTop
):
  GLSprite(w, h, mat, iLeft, iBottom, iRight, iTop)
{}

void GLFace2FaceSprite::render(GLCamera& camera) {
	glPushMatrix();
  glTranslatef(getPositionX(),getPositionY(),getPositionZ());
  M3Vector halfSide(
    camera.getSideDirectionX(),
    camera.getSideDirectionY(),
    camera.getSideDirectionZ()
  );
  halfSide.scale(getHalfWidth());
  M3Vector halfUp(
    camera.getUpDirectionX(),
    camera.getUpDirectionY(),
    camera.getUpDirectionZ()
  );
  halfUp.scale(getHalfHeight());
  M3Vector normal(
    -camera.getViewDirectionX(),
    -camera.getViewDirectionY(),
    -camera.getViewDirectionZ()
  );
  glDisable(GL_CULL_FACE);
  if(isTransparent()) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_FALSE);
  }
  glEnable(GL_ALPHA_TEST);
  material->applyColor();
  material->applyDiffuse();
  glBegin(GL_QUADS);
  glNormal3f(normal.getX(),normal.getY(),normal.getZ());
  glTexCoord2f(right,top);
  glVertex3f(
    -halfSide.getX()-halfUp.getX(),
    -halfSide.getY()-halfUp.getY(),
    -halfSide.getZ()-halfUp.getZ()
  );
  glTexCoord2f(right,bottom);
  glVertex3f(
    -halfSide.getX()+halfUp.getX(),
    -halfSide.getY()+halfUp.getY(),
    -halfSide.getZ()+halfUp.getZ()
  );
  glTexCoord2f(left,bottom);
  glVertex3f(
     halfSide.getX()+halfUp.getX(),
     halfSide.getY()+halfUp.getY(),
     halfSide.getZ()+halfUp.getZ()
  );
  glTexCoord2f(left,top);
  glVertex3f(
     halfSide.getX()-halfUp.getX(),
     halfSide.getY()-halfUp.getY(),
     halfSide.getZ()-halfUp.getZ()
  );
  glEnd();
  if(isTransparent()) {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
	}
  glDisable(GL_ALPHA_TEST);
  glEnable(GL_CULL_FACE);
	glPopMatrix();
}

//
// GLFace2FaceAnimatedSprite
//
GLFace2FaceAnimatedSprite::GLFace2FaceAnimatedSprite(
  float w, float h, float duration, int sideFrm, GLMaterial* mat
):
  GLAnimatedSprite(duration,sideFrm),
  GLSprite(w,h,mat,0,0,1.0f/sideFrm,1.0f/sideFrm)
{
}

void GLFace2FaceAnimatedSprite::render(GLCamera& camera) {
  if(isAnimated()) {
    glMatrixMode(GL_TEXTURE);
    glPushMatrix();
    glLoadIdentity();
    double integer;
    int frame = (int)(
      modf(
        GLWin::getElapsedTime()/loopDuration,&integer
      )*sideFrames*sideFrames
    );
    glTranslatef(frame%sideFrames*uvStep,frame/sideFrames*uvStep,0);
    glMatrixMode(GL_MODELVIEW);
  }
  GLFace2FaceSprite::render(camera);
  if(isAnimated()) {
    glMatrixMode(GL_TEXTURE);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
  }
}

//
// GLBillboard
//
GLBillboard::GLBillboard() {
}

GLBillboard::GLBillboard(
  float w, float h, GLMaterial* mat,
  float iLeft, float iBottom, float iRight, float iTop
):
  GLSprite(w, h, mat, iLeft, iBottom, iRight, iTop)
{}

void GLBillboard::render(GLCamera& camera) {
	glPushMatrix();
  glTranslatef(getPositionX(),getPositionY(),getPositionZ());
  M3Vector halfSide(camera.getSideDirectionX(),0,camera.getSideDirectionZ());
  halfSide.scale(getHalfWidth());
  glNormal3f(-camera.getViewXY().getX(),0,-camera.getViewXY().getZ());
  glDisable(GL_CULL_FACE);
  if(isTransparent()) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_FALSE);
  }
  glEnable(GL_ALPHA_TEST);
  material->applyColor();
  material->applyDiffuse();
  glBegin(GL_QUADS);
  glTexCoord2f(left,top);
  glVertex3f(-halfSide.getX(),-getHalfHeight(),-halfSide.getZ());
  glTexCoord2f(left,bottom);
  glVertex3f(-halfSide.getX(), getHalfHeight(),-halfSide.getZ());
  glTexCoord2f(right,bottom);
  glVertex3f( halfSide.getX(), getHalfHeight(), halfSide.getZ());
  glTexCoord2f(right,top);
  glVertex3f( halfSide.getX(),-getHalfHeight(), halfSide.getZ());
  glEnd();
  if(isTransparent()) {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
	}
  glDisable(GL_ALPHA_TEST);
  glEnable(GL_CULL_FACE);
	glPopMatrix();
}

//
// GLAnimatedBillboard
//
GLAnimatedBillboard::GLAnimatedBillboard(
  float w, float h, float duration, int sideFrm, GLMaterial* mat
):
  GLAnimatedSprite(duration,sideFrm),
  GLSprite(w,h,mat,0,0,1.0f/sideFrm,1.0f/sideFrm)
{}

void GLAnimatedBillboard::render(GLCamera& camera) {
  if(isAnimated()) {
    glMatrixMode(GL_TEXTURE);
    glPushMatrix();
    glLoadIdentity();
    double integer;
    int frame = (int)(
      modf(
        GLWin::getElapsedTime()/loopDuration,&integer
      )*sideFrames*sideFrames
    );
    glTranslatef(frame%sideFrames*uvStep,frame/sideFrames*uvStep,0);
    glMatrixMode(GL_MODELVIEW);
  }
  GLBillboard::render(camera);
  if(isAnimated()) {
    glMatrixMode(GL_TEXTURE);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
  }
};

//
// GLBillboardInfo
//
GLBillboardInfo::GLBillboardInfo(
  float x, float y, float z, float w, float h, int txtIdx, bool r, int matIdx
):
  GLSpriteRect(w,h), M3Vector(x,y,z),
  textureIndex(txtIdx), materialIndex(matIdx),
  reflected(r), distance2(0)
{}

//
// GLBillboards
//
GLBillboards::GLBillboards(int capacity, int sideTxt, GLMaterial* mat):
  billboards(capacity), materials(1), sideTextures(sideTxt)
{
  if(!mat) mat = new GLMaterial();
  addMaterial(mat->acquireReference());
}

void GLBillboards::addBillboard(
 float x, float y, float z, float w, float h, int txtIdx, bool refl, int matIdx
) {
  billboards.addElement((DSSortableObject*)
    new GLBillboardInfo(x,y,z,w,h,txtIdx,refl,matIdx)
  );
  float rad = x*x+y*y+z*z+w*w*0.5f+h*h*0.25f;
  if(rad > getMaxRadius()*getMaxRadius()) setMaxRadius(sqrt(rad));
}

void GLBillboards::render(GLCamera& camera) {
  for(int ct = 0; ct < billboards.getSize(); ct++) {
    GLBillboardInfo* bill = billboards.getElement(ct);
    M3Vector billPos(bill->getX(),bill->getY(),bill->getZ());
    billPos.multiply(getTransform());
    billPos.subtract(
      camera.getPositionX(),camera.getPositionY(),camera.getPositionZ()
    );
    bill->setDistance2(billPos.norm2());
  }
  billboards.quickSort(0,billboards.getSize()-1);
	glPushMatrix();
  glMultMatrixf(getTransform());
  M3Vector cameraSide(camera.getSideDirectionX(),0,camera.getSideDirectionZ());
  cameraSide.rotateT(getTransform());
  M3Vector cameraViewXY(camera.getViewXY().getX(),0,camera.getViewXY().getZ());
  cameraViewXY.rotateT(getTransform());
  const float textureStep = 1.0f/sideTextures;
  glDisable(GL_CULL_FACE);
  if(isTransparent()) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_FALSE);
  }
  glEnable(GL_ALPHA_TEST);
  glNormal3f(-cameraViewXY.getX(),0,-cameraViewXY.getZ());
  int oldMaterialIndex = -1;
  for(int ct = 0; ct < billboards.getSize(); ct++) {
    GLBillboardInfo* bill = billboards.getElement(ct);
    int materialIndex = bill->getMaterialIndex();
    if(materialIndex != oldMaterialIndex) {
      oldMaterialIndex = materialIndex;
      GLMaterial* material = materials.getElement(materialIndex);
      material->applyColor();
      material->applyDiffuse();
    }
    glBegin(GL_QUADS);
    M3Vector halfSide(cameraSide.getX(),0,cameraSide.getZ());
    halfSide.scale(bill->getHalfWidth());
    float u =
      bill->getTextureIndex()%sideTextures*textureStep+
      bill->isReflected()?textureStep:0;
    float uStep = bill->isReflected()? -textureStep: textureStep;
    float v = bill->getTextureIndex()/sideTextures*textureStep;
    glTexCoord2f(u,v+textureStep);
    glVertex3f(
      bill->getX()-halfSide.getX(),
      bill->getY()-bill->getHalfHeight(),
      bill->getZ()-halfSide.getZ()
    );
    glTexCoord2f(u,v);
    glVertex3f(
      bill->getX()-halfSide.getX(),
      bill->getY()+bill->getHalfHeight(),
      bill->getZ()-halfSide.getZ()
    );
    glTexCoord2f(u+uStep,v);
    glVertex3f(
      bill->getX()+halfSide.getX(),
      bill->getY()+bill->getHalfHeight(),
      bill->getZ()+halfSide.getZ()
    );
    glTexCoord2f(u+uStep,v+textureStep);
    glVertex3f(
      bill->getX()+halfSide.getX(),
      bill->getY()-bill->getHalfHeight(),
      bill->getZ()+halfSide.getZ()
    );
    glEnd();
  }
  if(isTransparent()) {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
  }
  glDisable(GL_ALPHA_TEST);
  glEnable(GL_CULL_FACE);
	glPopMatrix();
}

//
// GLTreeInfo
//
GLTreeInfo::GLTreeInfo(
  float x, float y, float z, float w, float h, int txtIdx, bool r, int matIdx
):
  GLBillboardInfo(x,y,z,w,h,txtIdx,r,matIdx), phase(0)
{}

//
// GLTrees
//
GLTrees::GLTrees(
  int capacity, int sideTxt, GLMaterial* mat, float wndDur, float wndDisp
):
  GLBillboards(capacity,sideTxt,mat),
  windDuration(wndDur), windDisplacement(wndDisp)
{}

void GLTrees::addTree(
 float x, float y, float z, float w, float h, int txtIdx, bool refl, int matIdx
) {
  GLTreeInfo* tree = new GLTreeInfo(x,y,z,w,h,txtIdx,refl,matIdx);
  tree->setPhase(windDuration*random(100)/100);
  billboards.addElement((DSSortableObject*)tree);
  float rad = x*x+y*y+z*z+w*w*0.5f+h*h*0.25f;
  if(rad > getMaxRadius()*getMaxRadius()) setMaxRadius(sqrt(rad));
}

void GLTrees::render(GLCamera& camera) {
  for(int ct = 0; ct < billboards.getSize(); ct++) {
    GLBillboardInfo* bill = billboards.getElement(ct);
    M3Vector billPos(bill->getX(),bill->getY(),bill->getZ());
    billPos.multiply(getTransform());
    billPos.subtract(
      camera.getPositionX(),camera.getPositionY(),camera.getPositionZ()
    );
    bill->setDistance2(billPos.norm2());
  }
  billboards.quickSort(0,billboards.getSize()-1);
	glPushMatrix();
  glMultMatrixf(getTransform());
  M3Vector cameraSide(camera.getSideDirectionX(),0,camera.getSideDirectionZ());
  cameraSide.rotateT(getTransform());
  M3Vector cameraViewXY(camera.getViewXY().getX(),0,camera.getViewXY().getZ());
  cameraViewXY.rotateT(getTransform());
  const float textureStep = 1.0f/sideTextures;
  glDisable(GL_CULL_FACE);
  if(isTransparent()) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_FALSE);
  }
  glEnable(GL_ALPHA_TEST);
  glNormal3f(-cameraViewXY.getX(),0,-cameraViewXY.getZ());
  int oldMaterialIndex = -1;
  for(int ct = 0; ct < billboards.getSize(); ct++) {
    GLTreeInfo* tree = (GLTreeInfo*)billboards.getElement(ct);
    float frac;
    if(windDuration != 0) {
      double integer;
      frac = (float)modf(
        (GLWin::getElapsedTime()+tree->getPhase())/windDuration,&integer
      );
      const float maxFrac = 0.667f;
      if(frac > maxFrac) frac = (1-frac)*(maxFrac/(1-maxFrac));
      frac *= windDisplacement*(1/maxFrac);
    }
    int materialIndex = tree->getMaterialIndex();
    if(materialIndex != oldMaterialIndex) {
      oldMaterialIndex = materialIndex;
      GLMaterial* material = materials.getElement(materialIndex);
      material->applyColor();
      material->applyDiffuse();
    }
    glBegin(GL_QUADS);
    M3Vector halfSide(cameraSide.getX(),0,cameraSide.getZ());
    halfSide.scale(tree->getHalfWidth());
    float u =
      tree->getTextureIndex()%sideTextures*textureStep+
      tree->isReflected()?textureStep:0;
    float uStep = tree->isReflected()? -textureStep: textureStep;
    float v = tree->getTextureIndex()/sideTextures*textureStep;
    glTexCoord2f(u,v+textureStep);
    glVertex3f(
      tree->getX()-halfSide.getX(),
      tree->getY()-tree->getHalfHeight(),
      tree->getZ()-halfSide.getZ()
    );
    if(windDuration != 0) {
      float fracX = frac*halfSide.getX();
      float fracZ = frac*halfSide.getZ();
      glTexCoord2f(u,v);
      glVertex3f(
        tree->getX()-halfSide.getX()+fracX,
        tree->getY()+tree->getHalfHeight(),
        tree->getZ()-halfSide.getZ()+fracZ
      );
      glTexCoord2f(u+uStep,v);
      glVertex3f(
        tree->getX()+halfSide.getX()+fracX,
        tree->getY()+tree->getHalfHeight(),
        tree->getZ()+halfSide.getZ()+fracZ
      );
    } else {
      glTexCoord2f(u,v);
      glVertex3f(
        tree->getX()-halfSide.getX(),
        tree->getY()+tree->getHalfHeight(),
        tree->getZ()-halfSide.getZ()
      );
      glTexCoord2f(u+uStep,v);
      glVertex3f(
        tree->getX()+halfSide.getX(),
        tree->getY()+tree->getHalfHeight(),
        tree->getZ()+halfSide.getZ()
      );
    }
    glTexCoord2f(u+uStep,v+textureStep);
    glVertex3f(
      tree->getX()+halfSide.getX(),
      tree->getY()-tree->getHalfHeight(),
      tree->getZ()+halfSide.getZ()
    );
    glEnd();
  }
  if(isTransparent()) {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
  }
  glDisable(GL_ALPHA_TEST);
  glEnable(GL_CULL_FACE);
	glPopMatrix();
}

