#ifndef GLSPRITE_H
#define GLSPRITE_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "globject.h"
#include "glmaterial.h"

#include <stdlib.h> // for random

//
// GLSpriteRect
//
class GLSpriteRect {
private:
  float halfWidth, halfHeight;
protected:
  GLSpriteRect();
public:
  GLSpriteRect(float w, float h);
  void setHalfWidth(float hw) {halfWidth = hw;}
  void setHalfHeight(float hh) {halfHeight = hh;}
  float getHalfWidth() {return halfWidth;}
  float getHalfHeight() {return halfHeight;}
};

//
// GLSprite
//
class GLSprite: public GLObject, public GLSpriteRect {
protected:
  GLMaterial* material;
  float left, bottom, right, top;
protected:
  GLSprite();
public:
  GLSprite(
    float w, float h, GLMaterial* mat = NULL,
    float iLeft = 0, float iBottom = 0, float iRight = 1, float iTop = 1
  );
  ~GLSprite();
  GLMaterial& getMaterial() {return *material;}
  void setMaterial(GLMaterial* mat) {
    if(material) material->releaseReference();
    material = mat->acquireReference();
  }
  virtual void render(GLCamera& camera);
};

//
// GLAnimatedSprite
//
class GLAnimatedSprite: virtual public GLSprite {
protected:
  float loopDuration;
  int sideFrames;
  bool animated;
  float uvStep;
protected:
  GLAnimatedSprite(float duration, float sideFrm = 1);
public:
  GLAnimatedSprite(
    float w, float h, float duration,
    float sideFrm = 1, GLMaterial* mat = NULL
  );
  void setLoopDuration(float ld) {loopDuration = ld;}
  float getLoopDuration() {return loopDuration;}
  bool isAnimated() {return animated;}
  void setAnimated(bool a) {animated = a;}
  virtual void render(GLCamera& camera);
};

//
// GLFace2FaceSprite
//
class GLFace2FaceSprite: virtual public GLSprite {
protected:
  GLFace2FaceSprite();
public:
  GLFace2FaceSprite(
    float w, float h, GLMaterial* mat = NULL,
    float iLeft = 0, float iBottom = 0, float iRight = 1, float iTop = 1
  );
  virtual void render(GLCamera& camera);
};

//
// GLFace2FaceAnimatedSprite
//
class GLFace2FaceAnimatedSprite:
  public GLFace2FaceSprite, public GLAnimatedSprite
{
public:
  GLFace2FaceAnimatedSprite(
    float w, float h, float duration, int sideFrm = 1, GLMaterial* mat = NULL
  );
  void render(GLCamera& camera);
};

//
// GLBillboard
//
class GLBillboard: virtual public GLSprite {
protected:
  GLBillboard();
public:
  GLBillboard(
    float w, float h, GLMaterial* mat = NULL,
    float iLeft = 0, float iBottom = 0, float iRight = 1, float iTop = 1
  );
  GLMaterial& getMaterial() {return *material;}
  void setMaterial(GLMaterial* mat) {
    if(material) material->releaseReference();
    material = mat->acquireReference();
  }
  virtual void render(GLCamera& camera);
};

//
// GLAnimatedBillboard
//
class GLAnimatedBillboard: public GLBillboard, public GLAnimatedSprite {
public:
  GLAnimatedBillboard(
    float w, float h, float duration, int sideFrm = 1, GLMaterial* mat = NULL
  );
  void render(GLCamera& camera);
};

//
// GLBillboardInfo
//
class GLBillboardInfo:
  public GLSpriteRect, public M3Vector, public DSSortableObject
{
protected:
  int materialIndex;
  int textureIndex;
  bool reflected;
  float distance2;
public:
  GLBillboardInfo(
    float x, float y, float z, float w, float h,
    int txtIdx = 0, bool refl = false, int matIdx = 0
  );
  void setMaterialIndex(int idx) {materialIndex = idx;}
  int getMaterialIndex() {return materialIndex;}
  void setTextureIndex(int idx) {textureIndex = idx;}
  int getTextureIndex() {return textureIndex;}
  void setReflected(bool r) {reflected = r;}
  bool isReflected() {return reflected;}
  void setDistance2(float d2) {distance2 = d2;}
  float getSortIndex() {return distance2;}
};

//
// GLBillboards
//
class GLBillboards: public GLObject {
protected:
  DSSortedArray<GLBillboardInfo> billboards;
  DSArray<GLMaterial> materials;
  int sideTextures;
public:
  GLBillboards(int capacity, int sideTxt = 1, GLMaterial* mat = NULL);
  void addBillboard(
    float x, float y, float z, float w, float h,
    int txtIdx = 0, bool refl = false, int matIdx = 0
  );
  GLMaterial* getMaterial(int idx) {return materials.getElement(idx);}
  void addMaterial(GLMaterial* mat)
    {materials.addElement(mat->acquireReference());}
  virtual void render(GLCamera& camera);
};

//
// GLTreeInfo
//
class GLTreeInfo: public GLBillboardInfo {
private:
  float phase;
public:
  GLTreeInfo(
    float x, float y, float z, float w, float h,
    int txtIdx = 0, bool refl = false, int matIdx = 0
  );
  void setPhase(float p) {phase = p;}
  float getPhase() {return phase;}
};

//
// GLTrees
//
class GLTrees: public GLBillboards {
private:
  float windDuration;
  float windDisplacement;
protected:
  void addBillboard(
    float x, float y, float z, float w, float h,
    int txtIdx = 0, bool refl = false, int matIdx = 0
  ) {GLBillboards::addBillboard(x,y,z,w,h,txtIdx,refl,matIdx);}
public:
  GLTrees(
    int capacity, int sideTxt = 1, GLMaterial* mat = NULL,
    float wndDur = 0, float wndDisp = 0.1f
  );
  void addTree(
    float x, float y, float z, float w, float h,
    int txtIdx = 0, bool refl = false, int matIdx = 0
  );
  void render(GLCamera& camera);
};

#endif // GLSPRITE_H
