/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glextensions.h"
#include "glterrain.h"

//
// GLFlatTerrain
//

GLFlatTerrain::GLFlatTerrain(GLMaterial* mat, float siz, int tils):
  material(mat->acquireReference()), visible(false), reflective(false),
  list(glGenLists(1)), size(siz), tiles(tils)
{
  const float INV_SIDE_QUADS_COUNT = 1.0f/SIDE_QUADS_COUNT;
  const float halfSize = size*0.5f;
  const float stepUV = (float)tiles*INV_SIDE_QUADS_COUNT;
  const float step = size*INV_SIDE_QUADS_COUNT;
  glNewList(list,GL_COMPILE);
  for(float y = -halfSize, v = 0; y < halfSize; y += step, v += stepUV) {
    glBegin(GL_TRIANGLE_STRIP);
    glNormal3f(0,1,0);
    for(float x = -halfSize, u = 0; x <= halfSize; x += step, u += stepUV) {
      glTexCoord2f(u,v);
      glVertex3f(x,0,y);
      glTexCoord2f(u,v+stepUV);
      glVertex3f(x,0,y+step);
    }
    glEnd();
  }
  glEndList();
}

GLFlatTerrain::~GLFlatTerrain() {
  material->releaseReference();
  glDeleteLists(list,1);
}

void GLFlatTerrain::render(GLCamera& camera) {
  glMatrixMode(GL_TEXTURE);
  glPushMatrix();
  glLoadIdentity();
  float tilesOverSize = tiles/size;
  glTranslatef(
    tilesOverSize*camera.getPositionX(),tilesOverSize*camera.getPositionZ(),0
  );
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glTranslatef(camera.getPositionX(),0,camera.getPositionZ());
  if(isReflective()) {
    glDisable(GL_CULL_FACE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE,GL_SRC_COLOR);
  }
  material->applyColor();
  if(material->hasDiffuse()) material->applyDiffuse();
  glCallList(list);
  glDisable(GL_TEXTURE_2D);
  glDepthMask(GL_FALSE);
  const float halfSize = size*0.5f;
  glBegin(GL_TRIANGLE_STRIP);
  glColor3f(1,1,1);
  glVertex3f(-halfSize,-1,-halfSize);
  glVertex3f(-halfSize,camera.getPositionY(),-halfSize);
  glVertex3f(-halfSize,-1, halfSize);
  glVertex3f(-halfSize,camera.getPositionY(), halfSize);
  glVertex3f( halfSize,-1, halfSize);
  glVertex3f( halfSize,camera.getPositionY(), halfSize);
  glVertex3f( halfSize,-1,-halfSize);
  glVertex3f( halfSize,camera.getPositionY(),-halfSize);
  glVertex3f(-halfSize,-1,-halfSize);
  glVertex3f(-halfSize,camera.getPositionY(),-halfSize);
  glEnd();
  glEnable(GL_TEXTURE_2D);
  glDepthMask(GL_TRUE);
  if(isReflective()) {
    glEnable(GL_CULL_FACE);
    glDisable(GL_BLEND);
  }
  glMatrixMode(GL_TEXTURE);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
}

//
// GLHeightField
//

GLHeightField::GLHeightField(
  DRImage* hm, GLMaterial* mat, float w, float d, float h, float wl, int detTils
):
  heightMap(hm), material(mat->acquireReference()),
  detailedTiles(detTils), list(-1), listCount(0)
{
  setSize(w,d,h,wl);
  compile();
}

GLHeightField::~GLHeightField() {
  glDeleteLists(list,listCount);
  material->releaseReference();
  if(heightMap) delete heightMap;
}

void GLHeightField::compile() {
  if(listCount > 0)
    glDeleteLists(list,listCount);
  listCount = waterLevel > 0? LODS_COUNT*2: LODS_COUNT;
  list = glGenLists(listCount);
  const float hScale = height*3.921569e-3f;
  const float halfWidth = width*0.5f;
  const float halfDepth = depth*0.5f;
  for(int ct = 0; ct < LODS_COUNT; ct++) {
    glNewList(list+ct,GL_COMPILE);
    int pow = 1<<ct;
    float stepUV = 1.0f/(heightMap->getHeight()-pow);
    float stepXX = width*stepUV;
    float stepYY = depth*stepUV;
    float stepXXpow = stepXX*pow;
    float stepYYpow = stepYY*pow;
    for(int y = 0; y < heightMap->getHeight()-pow; y += pow) {
      float yy = y*stepYY-halfDepth;
      float v = y*stepUV;
      bool verticesFound = false;
      int baseX = y*heightMap->getWidth();
      int baseX2 = baseX+pow*heightMap->getWidth();
      for(int x = 0; x < heightMap->getWidth()-pow; x += pow) {
        float xx = x*stepXX-halfWidth;
        float u = x*stepUV;
        int idx = baseX+x;
        float h0 = heightMap->getBuffer()[idx]*hScale;
        float h2 = heightMap->getBuffer()[baseX2+x]*hScale;
        if(h0 > waterLevel && h2 > waterLevel) {
          if(!verticesFound) {
            verticesFound = true;
            glBegin(GL_QUAD_STRIP);
          }
          float h1 = heightMap->getBuffer()[idx+pow]*hScale;
          M3Vector v1(0,h2-h0,stepYYpow);
          M3Vector v2(stepXXpow,h1-h0,0);
          v1.cross(v2).normalize();
          glNormal3f(v1.getX(),v1.getY(),v1.getZ());
          if(isMultiTexturingEnabled) {
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB,u,v);
            glMultiTexCoord2fARB(
              GL_TEXTURE1_ARB,u*detailedTiles,v*detailedTiles
            );
            glVertex3f(xx,h0,yy);
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB,u,v+stepUV);
            glMultiTexCoord2fARB(
              GL_TEXTURE1_ARB,u*detailedTiles,(v+stepUV)*detailedTiles
            );
            glVertex3f(xx,h2,yy+stepYYpow);
          } else {
            glTexCoord2f(u,v);
            glVertex3f(xx,h0,yy);
            glTexCoord2f(u,v+stepUV);
            glVertex3f(xx,h2,yy+stepYYpow);
          }
        } else {
          if(verticesFound) {
            verticesFound = false;
            glEnd();
          }
        }
      }
      if(verticesFound)
        glEnd();
    }
    glEndList();
  }
  if(waterLevel > 0) {
    for(int ct = 0; ct < LODS_COUNT; ct++) {
      glNewList(list+ct+LODS_COUNT,GL_COMPILE);
      int pow = 1<<ct;
      float stepUV = 1.0f/(heightMap->getHeight()-pow);
      float stepXX = width*stepUV;
      float stepYY = depth*stepUV;
      float stepXXpow = stepXX*pow;
      float stepYYpow = stepYY*pow;
      glBegin(GL_QUADS);
      for(int y = 0; y < heightMap->getHeight()-pow; y += pow) {
        float yy = y*stepYY-halfDepth;
        float v = y*stepUV;
        int baseX = y*heightMap->getWidth();
        int baseX2 = baseX+pow*heightMap->getWidth();
        for(int x = 0; x < heightMap->getWidth()-pow; x += pow) {
          float xx = x*stepXX-halfWidth;
          float u = x*stepUV;
          int idx = baseX+x;
          float h0 = heightMap->getBuffer()[idx]*hScale;
          float h1 = heightMap->getBuffer()[idx+pow]*hScale;
          int idx2 = baseX2+x;
          float h2 = heightMap->getBuffer()[idx2]*hScale;
          float h3 = heightMap->getBuffer()[idx2+pow]*hScale;
          if(
            (
              h0 > waterLevel &&
              (h1 < waterLevel || h2 < waterLevel || h3 < waterLevel)
            ) || (
              h0 < waterLevel &&
              (h1 > waterLevel || h2 > waterLevel || h3 > waterLevel)
            )
          ) {
            M3Vector v1(0,h2-h0,stepYYpow);
            M3Vector v2(stepXXpow,h1-h0,0);
            v1.cross(v2).normalize();
            glNormal3f(v1.getX(),v1.getY(),v1.getZ());
            if(isMultiTexturingEnabled) {
              glMultiTexCoord2fARB(GL_TEXTURE0_ARB,u,v);
              glMultiTexCoord2fARB(
                GL_TEXTURE1_ARB,u*detailedTiles,v*detailedTiles
              );
              glVertex3f(xx,h0,yy);
              glMultiTexCoord2fARB(GL_TEXTURE0_ARB,u,v+stepUV);
              glMultiTexCoord2fARB(
                GL_TEXTURE1_ARB,u*detailedTiles,(v+stepUV)*detailedTiles
              );
              glVertex3f(xx,h2,yy+stepYYpow);
              glMultiTexCoord2fARB(GL_TEXTURE0_ARB,u+stepUV,v+stepUV);
              glMultiTexCoord2fARB(
                GL_TEXTURE1_ARB,(u+stepUV)*detailedTiles,(v+stepUV)*detailedTiles
              );
              glVertex3f(xx+stepXXpow,h3,yy+stepYYpow);
              glMultiTexCoord2fARB(GL_TEXTURE0_ARB,u+stepUV,v);
              glMultiTexCoord2fARB(
                GL_TEXTURE1_ARB,(u+stepUV)*detailedTiles,v*detailedTiles
              );
              glVertex3f(xx+stepXXpow,h1,yy);
            } else {
              glTexCoord2f(u,v);
              glVertex3f(xx,h0,yy);
              glTexCoord2f(u,v+stepUV);
              glVertex3f(xx,h2,yy+stepYYpow);
              glTexCoord2f(u+stepUV,v+stepUV);
              glVertex3f(xx+stepXXpow,h3,yy+stepYYpow);
              glTexCoord2f(u+stepUV,v);
              glVertex3f(xx+stepXXpow,h1,yy);
            }
          }
        }
      }
      glEnd();
      glEndList();
    }
  }
}

void GLHeightField::setHeightMap(DRImage* hm) {
  if(heightMap) delete heightMap;
  heightMap = hm;
}

float GLHeightField::getHeightAtRelative(float x, float y) {
  x = x/width+0.5f;
  if(x < 0 || x > 1) return 0;
  x = x*heightMap->getWidth();
  int xx = (int)(x+0.5f);
  if(xx >= heightMap->getWidth()-1) return 0;
  x -= xx;
  y = y/depth+0.5f;
  if(y < 0 || y > 1) return 0;
  y = y*heightMap->getWidth();
  int yy = (int)(y+0.5f);
  if(yy >= heightMap->getHeight()-1) return 0;
  y -= yy;
  int idx = yy*heightMap->getWidth()+xx;
  float heightK = height*3.921569e-3f;
  float hx = heightK*heightMap->getBuffer()[idx+1]-waterLevel;
  float hy =
    heightK*heightMap->getBuffer()[idx+heightMap->getWidth()]-waterLevel;
  float h0;
  if(x+y < 1) {
    h0 = heightK*heightMap->getBuffer()[idx]-waterLevel;
    h0 += ((hx-h0)*x+(hy-h0)*y)*0.5f;
  } else {
    h0 =
      heightK*heightMap->getBuffer()[idx+heightMap->getWidth()+1]-waterLevel;
    h0 += ((h0-hx)*(1-x)+(h0-hy)*(1-y))*0.5f;
  }
  return h0;
}

float GLHeightField::getHeightAtAbsolute(float x, float y) {
  M3Vector pos(x,0,y);
  pos.subtract(getPositionX(),0,getPositionZ());
  pos.rotateT(getTransform());
  return getHeightAtRelative(pos.getX(),pos.getZ());
}

void GLHeightField::render(GLCamera& camera) {
  glPushMatrix();
  glMultMatrixf(getTransform());
  glDisable(GL_LIGHTING);
  float dx = getPositionX()-camera.getPositionX();
  float dy = getPositionY()-camera.getPositionY();
  float dz = getPositionZ()-camera.getPositionZ();
  float dist2 = dx*dx+dy*dy+dz*dz;
  float comp2 = maxRadius*maxRadius;
  LodMode lod =
    (dist2 < 4*comp2)? NEAR_LOD:
    (dist2 < 16*comp2)? NORMAL_LOD:
    (dist2 < 36*comp2)? FAR_LOD: FARTEST_LOD;
  material->applyColor();
  if(isMultiTexturingEnabled && material->hasGloss()) {
    glActiveTextureARB(GL_TEXTURE0_ARB);
    if(material->hasDiffuse()) material->applyDiffuse();
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    material->applyGloss();
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE_EXT);
    glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_EXT,GL_ADD_SIGNED_EXT);
  } else {
    if(material->hasDiffuse()) material->applyDiffuse();
  }
  glCallList(list+lod);
  if(waterLevel > 0) {
    if(camera.isRenderingReflections()) {
      double plane[4] = {0,1,0,-waterLevel};
      glClipPlane(GL_CLIP_PLANE0,plane);
      glEnable(GL_CLIP_PLANE0);
    }
    glCallList(list+lod+LODS_COUNT);
    if(camera.isRenderingReflections())
      glDisable(GL_CLIP_PLANE0);
  }
  if(isMultiTexturingEnabled && material->hasGloss()) {
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
  }
  glEnable(GL_LIGHTING);
  glPopMatrix();
}
