/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glwin.h"
#include "glconsole.h"
#include "fmsound.h"
#include "initdlg.h"

#include <stdlib.h> // for itoa()

//
// WndProc
//
LRESULT CALLBACK WndProc(
  HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam
) {
  static GLWin* glWin = NULL;
  switch(uMsg) {
    case WM_CREATE: {
      glWin = (GLWin*)(((LPCREATESTRUCT)lParam)->lpCreateParams);
      break;
    }
    case WM_ACTIVATE: {
      if(!HIWORD(wParam)) {
        glWin->setActive(TRUE);
      } else {
        glWin->setActive(FALSE);
      }
      break;
    }
    case WM_SYSCOMMAND: {
      switch(wParam) {
        case SC_SCREENSAVE:
        case SC_MONITORPOWER:
          return 0;
      }
      break;
    }
    case WM_CLOSE: {
      PostQuitMessage(0);
      return 0;
    }
    case WM_KEYDOWN: {
      glWin->keyDown(wParam);
      switch(wParam) {
        case VK_ESCAPE: {
          glWin->running = false;
          break;
        }
        case VK_F1: {
          switch(glWin->getHelpMode()) {
            default:
            case GLWin::REDUCED_HELP:
            case GLWin::HIDDEN_HELP:
            case GLWin::USER_HELP:
              glWin->setHelpMode(GLWin::FULL_HELP);
              break;
            case GLWin::FULL_HELP:
              glWin->setHelpMode(GLWin::HIDDEN_HELP);
              break;
          }
          break;
        }
        case VK_F2: {
          glWin->setShowFramerate(!glWin->isFramerateShown());
          break;
        }
        case VK_F3: {
          glWin->setShowConsole(!glWin->isConsoleShown());
          break;
        }
        case VK_F5: {
          GLConsole& console = GLConsole::get();
          if(console.isSuspended())
            console.resume();
          else
            console.suspend();
          break;
        }
        case VK_F9: {
          glWin->setVSyncOn(!glWin->isVSyncOn());
          glWin->applyVSync();
          break;
        }
        case VK_F10: {
          glWin->acquireMouse(!glWin->isMouseAcquired());
          ShowCursor(glWin->isMouseAcquired()?FALSE:TRUE);
          break;
        }
        default: {
          if(glWin->isConsoleShown()) {
            MSG msg;
            if(PeekMessage(&msg,NULL,0,0,PM_REMOVE)) {
              glWin->keyUp(wParam);
              TranslateMessage(&msg);
              DispatchMessage(&msg);
            }
          }
          break;
        }
      }
      return 0;
    }
    case WM_KEYUP: {
      glWin->keyUp(wParam);
      return 0;
    }
    case WM_CHAR: {
      if(glWin->isConsoleShown()) {
        GLConsole& console = GLConsole::get();
        if(
          !console.getScriptInterpreter().isRunning() ||
          console.isSuspended()
        )
          console.keyPressed((TCHAR)wParam)
        ;
      }
      break;
    }
    case WM_LBUTTONUP:
    case WM_MBUTTONUP:
    case WM_RBUTTONUP:
    case WM_LBUTTONDOWN:
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
    case WM_MOUSEMOVE: {
      if(glWin->isMouseAcquired()) {
        static bool centeringMouse = true;
        if(!centeringMouse) {
          glWin->mouseKeys = wParam;
          glWin->mouseDX += LOWORD(lParam)-glWin->centerX;
          glWin->mouseDY += glWin->centerY-HIWORD(lParam);
          glWin->centerMouse();
          centeringMouse = true;
        } else {
          centeringMouse = false;
        }
      }
      return 0;
    }
    case WM_SIZE:   {
      glWin->resizeWindow(LOWORD(lParam),HIWORD(lParam));
      return 0;
    }
    default:
      break;
  }
  return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

//
// GLWin
//

GLWin::GLWin() throw(Exception):
  running(true), mouseAcquired(false), textOverlayFont(NULL),
  helpMode(REDUCED_HELP), showFramerate(true), showConsole(false)
{
  LARGE_INTEGER queryTicksPerSec;
  QueryPerformanceFrequency(&queryTicksPerSec);
  ticksPeriod = 1.0f/queryTicksPerSec.QuadPart;
  FMSoundManager::initialize();
}

float GLWin::timeStep = 0;
float GLWin::invTimeStep = 0;
float GLWin::elapsedTime = 0;
float GLWin::ticksPeriod = 0;

GLWin::~GLWin() {
  FMSoundManager::finalize();
}

bool GLWin::createWindow() {
  GLuint   pixelFormat;
  WNDCLASS wc;
  DWORD    dwExStyle;
  DWORD    dwStyle;
  RECT     windowRect;
  windowRect.left   = (long)0;
  windowRect.right  = (long)getWidth();
  windowRect.top    = (long)0;
  windowRect.bottom = (long)getHeight();
  hInstance      = GetModuleHandle(NULL);
  wc.style       = CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
  wc.lpfnWndProc = (WNDPROC)WndProc;
  wc.cbClsExtra  = 0;
  wc.cbWndExtra  = sizeof(LONG);
  wc.hInstance   = hInstance;
  wc.hIcon       = LoadIcon(hInstance, MAKEINTRESOURCE(1));
  wc.hCursor     = LoadCursor(NULL,IDC_ARROW);
  wc.hbrBackground = NULL;
  wc.lpszMenuName  = NULL;
  wc.lpszClassName = "GLWin";
  if(!RegisterClass(&wc)) {
    showError("Failed To Register The Window Class");
    return FALSE;
  }
  if(isFullscreen()) {
    DEVMODE dmScreenSettings;
    memset(&dmScreenSettings,0,sizeof(dmScreenSettings));
    dmScreenSettings.dmSize = sizeof(dmScreenSettings);
    dmScreenSettings.dmPelsWidth    = getWidth();
    dmScreenSettings.dmPelsHeight   = getHeight();
    dmScreenSettings.dmBitsPerPel   = getColorDepth();
    dmScreenSettings.dmFields = DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;
    if(
      ChangeDisplaySettings(
        &dmScreenSettings,CDS_FULLSCREEN
      ) != DISP_CHANGE_SUCCESSFUL
    ) {
      if(
        askQuestion(
          "The requested fullscreen mode is not supported by\n"
          "your video card. Use windowed mode instead?"
        ) == IDYES
      ) {
        setFullscreen(false);
      } else {
        showError("Program Will Now Close");
        return FALSE;
      }
    }
  }
  if(isFullscreen()) {
    dwExStyle = WS_EX_APPWINDOW;
    dwStyle = WS_POPUP|WS_SYSMENU|WS_VISIBLE;
    ShowCursor(FALSE);
  } else {
    dwExStyle = WS_EX_APPWINDOW|WS_EX_WINDOWEDGE;
    dwStyle = WS_OVERLAPPEDWINDOW;
  }
  AdjustWindowRectEx(&windowRect,dwStyle,FALSE,dwExStyle);
  if((hWnd = CreateWindowEx(
    dwExStyle,
    "GLWin",
    getTitle(),
    dwStyle|WS_CLIPSIBLINGS|WS_CLIPCHILDREN,
    0, 0,
    windowRect.right-windowRect.left,
    windowRect.bottom-windowRect.top,
    NULL,
    NULL,
    hInstance,
    (LPVOID)this
  )) == NULL) {
    destroyWindow();
    showError("Window Creation Error");
    return FALSE;
  }
  PIXELFORMATDESCRIPTOR pfd = {
    sizeof(PIXELFORMATDESCRIPTOR),
    1,
    PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER,
    PFD_TYPE_RGBA,
    getColorDepth(),
    0, 0, 0, 0, 0, 0,
    0,
    0,
    0,
    0, 0, 0, 0,
    getDepthDepth(),
    getStencilDepth(),
    0,
    PFD_MAIN_PLANE,
    0,
    0, 0, 0
  };
  if((hDC = GetDC(hWnd)) == NULL) {
    shutdown("Can't Create A GL Device Context");
    return FALSE;
  }
  if((pixelFormat = ChoosePixelFormat(hDC,&pfd)) == NULL) {
    shutdown("Can't Find A Suitable PixelFormat");
    return FALSE;
  }
  if(!SetPixelFormat(hDC,pixelFormat,&pfd)) {
    shutdown("Can't Set The PixelFormat");
    return FALSE;
  }
  if((hRC = wglCreateContext(hDC)) == NULL) {
    shutdown("Can't Create A GL Rendering Context");
    return FALSE;
  }
  if(!wglMakeCurrent(hDC,hRC)) {
    shutdown("Can't Activate The GL Rendering Context");
    return FALSE;
  }
  if(!initializeRenderer()) {
    shutdown("Can't Initialize The GL Rendering");
    return FALSE;
  }
  if(!initializeScene()) {
    shutdown("Can't Initialize The Scene");
    return FALSE;
  }
  ShowWindow(hWnd,SW_SHOW);
  SetForegroundWindow(hWnd);
  SetFocus(hWnd);
  resizeWindow(getWidth(),getHeight());
  if(!getTextOverlayFont())
    setTextOverlayFont(new GLBitmapFont(hDC,"Arial",16,0,FW_BOLD));
  acquireMouse(true);
  return TRUE;
}

void GLWin::destroyWindow() {
  finalizeScene();
  finalizeRenderer();
  if(fullscreen) {
    ChangeDisplaySettings(NULL,0);
    ShowCursor(TRUE);
  }
  if(hRC) {
    if(!wglMakeCurrent(NULL,NULL)) {
      showError("Release Of DC And RC Failed");
    }
    if(!wglDeleteContext(hRC)) {
      showError("Release Rendering Context Failed");
    }
    hRC = NULL;
  }
  if(hDC && !ReleaseDC(hWnd,hDC)) {
    showError("Release Device Context Failed");
    hDC = NULL;
  }
  if(hWnd && !DestroyWindow(hWnd)) {
    showError("Could Not Release The Window");
    hWnd = NULL;
  }
  if(!UnregisterClass("GLWin",hInstance)) {
    //showError("Could Not Unregister Class");
    hInstance = NULL;
  }
  if(getTextOverlayFont()) setTextOverlayFont(NULL);
}

void GLWin::resizeWindow(GLsizei w, GLsizei h) {
  if(h == 0) h = 1;
  setSize(w,h);
  centerX = w>>1;
  centerY = h>>1;
  if(isMouseAcquired())
    centerMouse();
  glViewport(0,0,w,h);
  updatePerspective();
}

void GLWin::drawTextOverlay() {
  GLFont* font = getTextOverlayFont();
  font->initialize();
  const int FONT_HEIGHT = font->getHeight();
  const int BASE_ROW = FONT_HEIGHT>>1;
  bool isFontTextured = font->isTextured();
  if(isFontTextured) {
    ((GLTexturedFont*)font)->applyTexture();
    glEnable(GL_ALPHA_TEST);
  } else {
    glDisable(GL_TEXTURE_2D);
  }
  if(getHelpMode() != HIDDEN_HELP) {
    if(getHelpMode() == REDUCED_HELP) {
      font->setColor(1,1,0);
      font->drawText(
        BASE_ROW,BASE_ROW+FONT_HEIGHT<<1,20,"[F1]  Show/Hide Help"
      );
    } else {
      int row = BASE_ROW+FONT_HEIGHT;
      if(getHelpMode() == FULL_HELP) {
        const char* helpTextLines[] = {
          "[F11] Fullscreen/Windowed",
          "[F10] Release/Acquire Mouse",
          "[F9]  On/Off Vertical Sync",
          "[F5]  Suspend/Resume Script",
          "[F3]  Show/Hide Console",
          "[F2]  Show/Hide FPS",
          "[F1]  Show/Hide Help",
          "[ESC] Quit",
          NULL
        };
        font->setColor(1,1,0);
        for(int ct = 0; helpTextLines[ct] != NULL; ct++) {
          int len = strlen(helpTextLines[ct]);
          row += FONT_HEIGHT;
          font->drawText(BASE_ROW,row,len,helpTextLines[ct]);
        }
        row += FONT_HEIGHT;
      }
      char* userHelpLine;
      font->setColor(1,1,1);
      for(int ct = 0; (userHelpLine = getHelpLine(ct)) != NULL; ct++) {
        row += FONT_HEIGHT;
        font->drawText(BASE_ROW,row,strlen(userHelpLine),userHelpLine);
      }
    }
  }
  if(isFramerateShown()) {
    static char fpsStr[16] = "";
    static int fpsLen = 0;
    static int frames = 0;
    frames++;
    static float startTime = GLWin::getElapsedTime();
    if(GLWin::getElapsedTime()-startTime >= 2) {
      strcat(itoa(frames>>1,fpsStr,10)," fps");
      fpsLen = strlen(fpsStr);
      frames = 0;
      startTime = GLWin::getElapsedTime();
    }
    font->setColor(1,1,0);
    font->drawText(BASE_ROW,BASE_ROW,fpsLen,fpsStr);
  }
  if(isConsoleShown()) {
    GLConsole& console = GLConsole::get();
    const int FONT_WIDTH = font->getSpacing();
    const int BASE_COL = FONT_WIDTH>>1;
    int rows = console.getRows();
    int consoleHeight = FONT_HEIGHT*rows;
    int consoleWidth = FONT_WIDTH*console.getCols();
    glLoadIdentity();
    glColor3f(0,0,0.2f);
    glEnable(GL_BLEND);
    if(isFontTextured) {
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_ALPHA_TEST);
    }
    glBlendFunc(GL_ONE,GL_ONE_MINUS_DST_COLOR);
    glBegin(GL_QUADS);
    glVertex2i(getWidth()-BASE_COL*2,BASE_ROW);
    glVertex2i(getWidth()-BASE_COL*2,BASE_ROW*3+consoleHeight);
    glVertex2i(getWidth()-BASE_COL*4-consoleWidth,BASE_ROW*3+consoleHeight);
    glVertex2i(getWidth()-BASE_COL*4-consoleWidth,BASE_ROW);
    glEnd();
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    if(isFontTextured) {
      glEnable(GL_TEXTURE_2D);
      glEnable(GL_ALPHA_TEST);
    }
    glDisable(GL_BLEND);
    int baseHorPosition = getWidth()-consoleWidth-FONT_WIDTH-2*BASE_COL;
    font->setColor(1,1,1);
    for(int ct = 0; ct < rows; ct++) {
      char* row = console.getRow(ct);
      font->drawText(
        baseHorPosition,2*BASE_ROW+(rows-ct-1)*FONT_HEIGHT,strlen(row),row
      );
    }
  }
  font->flushText();
  if(isFontTextured) {
    glDisable(GL_ALPHA_TEST);
  } else {
    glEnable(GL_TEXTURE_2D);
  }
}

//
// WinMain
//

int WINAPI WinMain(
  HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow
) {
  hPrevInstance, lpCmdLine, nCmdShow;
  GLWin* glWin;
  try {
    glWin = GLWin::newInstance(lpCmdLine);
  } catch(Exception& e) {
    return FALSE;
  }
  HRESULT hr = DialogBoxParam(
    hInstance,MAKEINTRESOURCE(1),NULL,(DLGPROC)InitDlgProc,(LPARAM)glWin
  );
  int retVal = FALSE;
  if(hr == IDCANCEL) GLWin::showError("Configuration cancelled");
  else if(hr < 0) GLWin::showError("Error during configuration");
  else if(glWin->createWindow()) {
    MSG msg;
    while(glWin->isRunning()) {
      if(PeekMessage(&msg,NULL,0,0,PM_REMOVE)) {
        if(msg.message == WM_QUIT) {
          break;
        } else {
          TranslateMessage(&msg);
          DispatchMessage(&msg);
        }
      } else {
        if(glWin->isActive()) {
          if(glWin->isKeyPressed(VK_F11)) {
            glWin->keyUp(VK_F11);
            glWin->destroyWindow();
            glWin->setFullscreen(!glWin->isFullscreen());
            glWin->setSize(glWin->getFullWidth(),glWin->getFullHeight());
            if(!glWin->createWindow())
              return FALSE;
          } else {
            LARGE_INTEGER queryTicks;
            QueryPerformanceCounter(&queryTicks);
            static LONGLONG lastTicks = queryTicks.QuadPart;
            GLWin::timeStep = GLWin::ticksPeriod*(
              queryTicks.QuadPart-lastTicks
            );
            GLWin::invTimeStep = 1/GLWin::timeStep;
            GLWin::elapsedTime += GLWin::timeStep;
            lastTicks = queryTicks.QuadPart;
            GLConsole& console = GLConsole::get();
            if(!console.isSuspended())
              console.continueExecution(GLWin::timeStep);
            glWin->updateScene();
            glWin->executeRendering();
            FMSoundManager::update3DSound();
            SwapBuffers(glWin->getHDC());
          }
        }
      }
    }
    retVal = msg.wParam;
  }
  glWin->destroyWindow();
  delete glWin;
  return retVal;
}

