/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include "glworld.h"

#include "datasets.cpp"

//
// GLWorld
//

GLWorld::GLWorld():
  backgrounds(2), transparentObjects(16), opaqueObjects(32), lights(8),
  sun(NULL), flatTerrain(NULL)
{}

GLWorld::~GLWorld() {
  removeEnvironment();
}

void GLWorld::removeEnvironment() {
  backgrounds.deleteElements();
  if(flatTerrain) {
    delete flatTerrain;
    flatTerrain = NULL;
  }
  if(sun) {
    delete sun;
    sun = NULL;
  }
}

void GLWorld::setBackground(GLBackground* bg) {
  backgrounds.deleteElements();
  addBackground(bg);
}

void GLWorld::setSun(GLSunLight* s) {
  if(sun) delete sun;
  sun = s;
}

void GLWorld::setFlatTerrain(GLFlatTerrain* t) {
  if(flatTerrain) delete flatTerrain;
  flatTerrain = t;
}

void GLWorld::enableFog(float dist, GLColor& clr) {
  glEnable(GL_FOG);
  glFogi(GL_FOG_MODE,GL_EXP2);
  glFogfv(GL_FOG_COLOR,clr.getFloatArray());
  glFogf(GL_FOG_DENSITY,sqrt(4.60517f/(dist*dist)));
}

void GLWorld::render(GLCamera& camera) {
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,ambient.getFloatArray());
  if(flatTerrain && flatTerrain->isVisible() && flatTerrain->isReflective())
    camera.setRenderingReflections(true);
  for(int ct = 0; ct < lights.getSize(); ct++) {
    GLObject* lgh = getLight(ct);
    if(lgh->isVisible()) lgh->clip(camera);
  }
  lights.quickSort(0,lights.getSize()-1);
  for(int ct = 0; ct < 8; ct++)
    glDisable(GL_LIGHT0+ct);
  for(int ct = 0; ct < getOpaqueObjectsCount(); ct++) {
    GLObject* obj = getOpaqueObject(ct);
    if(obj->isVisible()) obj->clip(camera);
  }
  for(int ct = 0; ct < getTransparentObjectsCount(); ct++) {
    GLObject* obj = getTransparentObject(ct);
    if(obj->isVisible()) obj->clip(camera);
  }
  opaqueObjects.quickSort(0,getOpaqueObjectsCount()-1);
  transparentObjects.quickSort(0,getTransparentObjectsCount()-1);
  soundSources.quickSort(0,getSoundSourcesCount()-1);
  for(int ct = getSoundSourcesCount()-1; ct >= 0; ct--) {
    GLSoundSource* src = getSoundSource(ct);
    FM3DSound* snd = src->get3DSound();
    switch(snd->getSoundStatus()) {
      case FMSound::IS_SOUND_PLAYING:
        snd->setPriority(ct);
        src->render(camera);
        break;
      case FMSound::IS_CHANNEL_LOST:
        snd->play(ct);
        if(snd->isAudible())
          src->render(camera);
        break;
      case FMSound::WAS_SOUND_STOPPED:
      default:
        break;
    }
  }
  if(backgrounds.getSize() > 0) {
    camera.applyRot();
    backgrounds.getElement(0)->render();
    for(int ct = 1; ct < backgrounds.getSize(); ct++)
      backgrounds.getElement(ct)->render(true);
  }
  camera.apply();
  if(flatTerrain && flatTerrain->isVisible() && flatTerrain->isReflective()) {
    glPushMatrix();
    glScalef(1,-1,1);
  }
  int enabledLights = 0;
  if(sun && sun->isVisible()) {
    M3Vector sunPosition(sun->getDirection());
    sunPosition.scale(sun->getCameraDistance()).add(
      camera.getPositionX(),camera.getPositionY(),camera.getPositionZ()
    );
    sun->setPosition(sunPosition);
    sun->clip(camera);
    sun->applyPropertiesTo(enabledLights++);
  }
  bool nearestLightFound = false;
  for(int ct = lights.getSize()-1; (ct >= 0) && (enabledLights < 8); ct--) {
    GLLight* light = (GLLight*)(getLight(ct));
    if(light->isVisible()) {
      if(!nearestLightFound) {
        bool isDirectional = light->isDirectional();
        camera.setNearestLight(isDirectional,
          isDirectional? light->getDirection(): light->getPosition()
        );
        nearestLightFound = true;
      }
      light->applyPropertiesTo(enabledLights++);
    }
  }
  if(!nearestLightFound && sun && sun->isVisible())
    camera.setNearestLight(true,sun->getDirection());
  if(flatTerrain && flatTerrain->isVisible() && flatTerrain->isReflective()) {
    glCullFace(GL_FRONT);
    for(int ct = getOpaqueObjectsCount()-1; ct >= 0; ct--) {
      GLObject* obj = getOpaqueObject(ct);
      if((!obj->isClipped()) && (obj->isVisible()))
        obj->render(camera);
    }
    for(int ct = 0; ct < getTransparentObjectsCount(); ct++) {
      GLObject* obj = getTransparentObject(ct);
      if((!obj->isClipped()) && (obj->isVisible()))
        obj->render(camera);
    }
    glCullFace(GL_BACK);
    glPopMatrix();
    camera.setRenderingReflections(false);
    int index = 0;
    if(sun && sun->isVisible())
      sun->applyPositionTo(index++);
    for(int ct = lights.getSize()-1; ct >= 0; ct--) {
      GLLight* light = (GLLight*)(getLight(ct));
      if(light->isVisible())
        light->applyPositionTo(index++);
    }
  }
  if(flatTerrain && flatTerrain->isVisible())
    flatTerrain->render(camera);
  for(int ct = getOpaqueObjectsCount()-1; ct >= 0; ct--) {
    GLObject* obj = getOpaqueObject(ct);
    if((!obj->isClipped()) && (obj->isVisible()))
      obj->render(camera);
  }
  for(int ct = 0; ct < getTransparentObjectsCount(); ct++) {
    GLObject* obj = getTransparentObject(ct);
    if((!obj->isClipped()) && (obj->isVisible()))
      obj->render(camera);
  }
  for(int ct = 0; ct < lights.getSize(); ct++) {
    GLObject* light = getLight(ct);
    if((!light->isClipped()) && (light->isVisible()))
      light->render(camera);
  }
  if(sun && (!sun->isClipped()) && sun->isVisible())
    sun->render(camera);
}

