#ifndef MATH3D_H
#define MATH3D_H

/*
  Copyright (C) 2001-2002 Leonardo Boselli

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Please send remarks, questions and bug reports to
    boselli@uno.it
  or write to:
    Leonardo Boselli
    Via Paoletti, 33
    18100 Imperia
    ITALY
*/

#include <math.h>

//
// M3Float
//
typedef float M3Float;

//
// M3Vector
//
class M3Vector {
private:
  enum {X = 0, Y, Z};
  M3Float vector[3];
public:
  M3Vector(M3Float x = 0, M3Float y = 0, M3Float z = 0);
  M3Vector(M3Float* v);
  M3Vector(M3Vector& v);
  M3Float getX() const {return vector[X];}
  M3Float getY() const {return vector[Y];}
  M3Float getZ() const {return vector[Z];}
  M3Float get(int index) {return vector[index];}
  void setX(M3Float x) {vector[X] = x;}
  void setY(M3Float y) {vector[Y] = y;}
  void setZ(M3Float z) {vector[Z] = z;}
  void set(int index, M3Float value) {vector[index] = value;}
  M3Vector& set(M3Float x = 0, M3Float y = 0, M3Float z = 0)
    {setX(x); setY(y); setZ(z); return *this;}
  M3Vector& set(M3Float* v)
    {setX(v[X]); setY(v[Y]); setZ(v[Z]); return *this;}
  M3Vector& set(M3Vector& v)
    {return set(v.getX(),v.getY(),v.getZ());}
  M3Vector& add(M3Float x, M3Float y, M3Float z)
    {return set(getX()+x,getY()+y,getZ()+z);}
  M3Vector& add(M3Float* v)
    {return set(getX()+v[X],getY()+v[Y],getZ()+v[Z]);}
  M3Vector& add(M3Vector& v)
    {return add(v.getX(),v.getY(),v.getZ());}
  M3Vector& add(M3Vector& v1, M3Vector& v2)
    {return set(v1.getX()+v2.getX(),v1.getY()+v2.getY(),v1.getZ()+v2.getZ());}
  M3Vector& subtract(M3Float x, M3Float y, M3Float z)
    {return set(getX()-x,getY()-y,getZ()-z);}
  M3Vector& subtract(M3Float* v)
    {return set(getX()-v[X],getY()-v[Y],getZ()-v[Z]);}
  M3Vector& subtract(M3Vector& v)
    {return subtract(v.getX(),v.getY(),v.getZ());}
  M3Vector& subtract(M3Vector& v1, M3Vector& v2)
    {return set(v1.getX()-v2.getX(),v1.getY()-v2.getY(),v1.getZ()-v2.getZ());}
  M3Vector& scale(M3Float s)
    {return set(s*getX(),s*getY(),s*getZ());}
  M3Float dot(M3Vector& v)
    {return getX()*v.getX()+getY()*v.getY()+getZ()*v.getZ();}
  M3Vector& cross(M3Vector& v) {return cross(*this,v);}
  M3Vector& cross(M3Vector& v1, M3Vector& v2) {
    return set(
      v1.getY()*v2.getZ()-v1.getZ()*v2.getY(),
      v1.getZ()*v2.getX()-v1.getX()*v2.getZ(),
      v1.getX()*v2.getY()-v1.getY()*v2.getX()
    );
  }
  void plane(M3Vector& v1, M3Vector& v2) {
    v1.set(1,1,-(getX()+getY())/getZ()).normalize();
    v2.cross(v1,*this).normalize();
  }
  M3Float norm2() {return getX()*getX()+getY()*getY()+getZ()*getZ();}
  M3Float norm() {return sqrt(norm2());}
  M3Vector& normalize() {return scale(1/norm());}
  M3Vector& multiply(M3Float* m) {
    return set(
      getX()*m[4*0+0]+getY()*m[4*1+0]+getZ()*m[4*2+0]+m[4*3+0],
      getX()*m[4*0+1]+getY()*m[4*1+1]+getZ()*m[4*2+1]+m[4*3+1],
      getX()*m[4*0+2]+getY()*m[4*1+2]+getZ()*m[4*2+2]+m[4*3+2]
    );
  }
  M3Vector& rotate(M3Float* m) {
    return set(
      getX()*m[4*0+0]+getY()*m[4*1+0]+getZ()*m[4*2+0],
      getX()*m[4*0+1]+getY()*m[4*1+1]+getZ()*m[4*2+1],
      getX()*m[4*0+2]+getY()*m[4*1+2]+getZ()*m[4*2+2]
    );
  }
  M3Vector& rotateT(M3Float* m) {
    return set(
      getX()*m[4*0+0]+getY()*m[4*0+1]+getZ()*m[4*0+2],
      getX()*m[4*1+0]+getY()*m[4*1+1]+getZ()*m[4*1+2],
      getX()*m[4*2+0]+getY()*m[4*2+1]+getZ()*m[4*2+2]
    );
  }
  M3Float* getFloatArray() {return vector;}
};

//
// M3Matrix
//
class M3Matrix {
protected:
  M3Float matrix[16];
public:
  M3Matrix(
    M3Float m11 = 1, M3Float m12 = 0, M3Float m13 = 0, M3Float m14 = 0,
    M3Float m21 = 0, M3Float m22 = 1, M3Float m23 = 0, M3Float m24 = 0,
    M3Float m31 = 0, M3Float m32 = 0, M3Float m33 = 1, M3Float m34 = 0,
    M3Float m41 = 0, M3Float m42 = 0, M3Float m43 = 0, M3Float m44 = 1
  );
  M3Matrix(M3Matrix& m);
  M3Matrix(M3Vector& v1, M3Vector& v2, M3Vector& v3);
  M3Matrix& set(
    M3Float m11 = 1, M3Float m12 = 0, M3Float m13 = 0, M3Float m14 = 0,
    M3Float m21 = 0, M3Float m22 = 1, M3Float m23 = 0, M3Float m24 = 0,
    M3Float m31 = 0, M3Float m32 = 0, M3Float m33 = 1, M3Float m34 = 0,
    M3Float m41 = 0, M3Float m42 = 0, M3Float m43 = 0, M3Float m44 = 1
  ) {
    set11(m11); set12(m12); set13(m13); set14(m14);
    set21(m21); set22(m22); set23(m23); set24(m24);
    set31(m31); set32(m32); set33(m33); set34(m34);
    set41(m41); set42(m42); set43(m43); set44(m44);
    return *this;
  }
  M3Matrix& set(M3Matrix& m) {
    return set(
      m.get11(), m.get12(), m.get13(), m.get14(),
      m.get21(), m.get22(), m.get23(), m.get24(),
      m.get31(), m.get32(), m.get33(), m.get34(),
      m.get41(), m.get42(), m.get43(), m.get44()
    );
  }
  M3Float get(int idx) {return matrix[idx];}
  M3Float get11() {return matrix[0*4+0];}
  M3Float get12() {return matrix[0*4+1];}
  M3Float get13() {return matrix[0*4+2];}
  M3Float get14() {return matrix[0*4+3];}
  M3Float get21() {return matrix[1*4+0];}
  M3Float get22() {return matrix[1*4+1];}
  M3Float get23() {return matrix[1*4+2];}
  M3Float get24() {return matrix[1*4+3];}
  M3Float get31() {return matrix[2*4+0];}
  M3Float get32() {return matrix[2*4+1];}
  M3Float get33() {return matrix[2*4+2];}
  M3Float get34() {return matrix[2*4+3];}
  M3Float get41() {return matrix[3*4+0];}
  M3Float get42() {return matrix[3*4+1];}
  M3Float get43() {return matrix[3*4+2];}
  M3Float get44() {return matrix[3*4+3];}
  void set(int idx, M3Float f) {matrix[idx] = f;}
  void set11(M3Float f) {matrix[0*4+0] = f;}
  void set12(M3Float f) {matrix[0*4+1] = f;}
  void set13(M3Float f) {matrix[0*4+2] = f;}
  void set14(M3Float f) {matrix[0*4+3] = f;}
  void set21(M3Float f) {matrix[1*4+0] = f;}
  void set22(M3Float f) {matrix[1*4+1] = f;}
  void set23(M3Float f) {matrix[1*4+2] = f;}
  void set24(M3Float f) {matrix[1*4+3] = f;}
  void set31(M3Float f) {matrix[2*4+0] = f;}
  void set32(M3Float f) {matrix[2*4+1] = f;}
  void set33(M3Float f) {matrix[2*4+2] = f;}
  void set34(M3Float f) {matrix[2*4+3] = f;}
  void set41(M3Float f) {matrix[3*4+0] = f;}
  void set42(M3Float f) {matrix[3*4+1] = f;}
  void set43(M3Float f) {matrix[3*4+2] = f;}
  void set44(M3Float f) {matrix[3*4+3] = f;}
  M3Float* multiply(M3Float* vec) {
    M3Float x = vec[0]*get11()+vec[1]*get21()+vec[2]*get31()+get41();
    M3Float y = vec[0]*get12()+vec[1]*get22()+vec[2]*get32()+get42();
    M3Float z = vec[0]*get13()+vec[1]*get23()+vec[2]*get33()+get43();
    vec[0] = x; vec[1] = y; vec[2] = z;
    return vec;
  }
  M3Vector& multiply(M3Vector& v) {
    return v.set(
      v.getX()*get11()+v.getY()*get21()+v.getZ()*get31()+get41(),
      v.getX()*get12()+v.getY()*get22()+v.getZ()*get32()+get42(),
      v.getX()*get13()+v.getY()*get23()+v.getZ()*get33()+get43()
    );
  }
  M3Vector& multiply(M3Vector& res, M3Vector& vec)
    {res = vec; return multiply(res);}
  M3Matrix& multiply(M3Matrix& mat) {return multiply(*this,mat);}
  M3Matrix& multiply(M3Matrix& res, M3Matrix& mat) {
    return res.set(
      res.get11()*mat.get11()+res.get12()*mat.get21()+res.get13()*mat.get31(),
      res.get11()*mat.get12()+res.get12()*mat.get22()+res.get13()*mat.get32(),
      res.get11()*mat.get13()+res.get12()*mat.get23()+res.get13()*mat.get33(),
      0,
      res.get21()*mat.get11()+res.get22()*mat.get21()+res.get23()*mat.get31(),
      res.get21()*mat.get12()+res.get22()*mat.get22()+res.get23()*mat.get32(),
      res.get21()*mat.get13()+res.get22()*mat.get23()+res.get23()*mat.get33(),
      0,
      res.get31()*mat.get11()+res.get32()*mat.get21()+res.get33()*mat.get31(),
      res.get31()*mat.get12()+res.get32()*mat.get22()+res.get33()*mat.get32(),
      res.get31()*mat.get13()+res.get32()*mat.get23()+res.get33()*mat.get33(),
      0,
      res.get41()*mat.get11()+res.get42()*mat.get21()+res.get43()*mat.get31()
      +mat.get41(),
      res.get41()*mat.get12()+res.get42()*mat.get22()+res.get43()*mat.get32()
      +mat.get42(),
      res.get41()*mat.get13()+res.get42()*mat.get23()+res.get43()*mat.get33()
      +mat.get43(),
      1
    );
  }
  M3Float* rotate(M3Float* vec) {
    M3Float x = vec[0]*get11()+vec[1]*get21()+vec[2]*get31();
    M3Float y = vec[0]*get12()+vec[1]*get22()+vec[2]*get32();
    M3Float z = vec[0]*get13()+vec[1]*get23()+vec[2]*get33();
    vec[0] = x; vec[1] = y; vec[2] = z;
    return vec;
  }
  M3Matrix& rotate(M3Matrix& mat) {return rotate(*this,mat);}
  M3Matrix& rotate(M3Matrix& res, M3Matrix& mat) {
    return res.set(
      res.get11()*mat.get11()+res.get12()*mat.get21()+res.get13()*mat.get31(),
      res.get11()*mat.get12()+res.get12()*mat.get22()+res.get13()*mat.get32(),
      res.get11()*mat.get13()+res.get12()*mat.get23()+res.get13()*mat.get33(),
      0,
      res.get21()*mat.get11()+res.get22()*mat.get21()+res.get23()*mat.get31(),
      res.get21()*mat.get12()+res.get22()*mat.get22()+res.get23()*mat.get32(),
      res.get21()*mat.get13()+res.get22()*mat.get23()+res.get23()*mat.get33(),
      0,
      res.get31()*mat.get11()+res.get32()*mat.get21()+res.get33()*mat.get31(),
      res.get31()*mat.get12()+res.get32()*mat.get22()+res.get33()*mat.get32(),
      res.get31()*mat.get13()+res.get32()*mat.get23()+res.get33()*mat.get33(),
      0,
      res.get41(),res.get42(),res.get43(),1
    );
  }
  M3Matrix& rotateC(M3Matrix& mat) {return rotateC(*this,mat);}
  M3Matrix& rotateC(M3Matrix& res, M3Matrix& mat) {
    return res.set(
      mat.get11()*res.get11()+mat.get12()*res.get21()+mat.get13()*res.get31(),
      mat.get11()*res.get12()+mat.get12()*res.get22()+mat.get13()*res.get32(),
      mat.get11()*res.get13()+mat.get12()*res.get23()+mat.get13()*res.get33(),
      0,
      mat.get21()*res.get11()+mat.get22()*res.get21()+mat.get23()*res.get31(),
      mat.get21()*res.get12()+mat.get22()*res.get22()+mat.get23()*res.get32(),
      mat.get21()*res.get13()+mat.get22()*res.get23()+mat.get23()*res.get33(),
      0,
      mat.get31()*res.get11()+mat.get32()*res.get21()+mat.get33()*res.get31(),
      mat.get31()*res.get12()+mat.get32()*res.get22()+mat.get33()*res.get32(),
      mat.get31()*res.get13()+mat.get32()*res.get23()+mat.get33()*res.get33(),
      0,
      res.get41(),res.get42(),res.get43(),1
    );
  }
  M3Float* getFloatArray() {return matrix;}
};

//
// M3RotationX
//
class M3RotationX: public M3Matrix {
public:
  M3RotationX(M3Float angle);
};

//
// M3RotationY
//
class M3RotationY: public M3Matrix {
public:
  M3RotationY(M3Float angle);
};

//
// M3RotationZ
//
class M3RotationZ: public M3Matrix {
public:
  M3RotationZ(M3Float angle);
};

//
// M3Reference
//
class M3Reference: public M3Matrix {
protected:
  enum {
    SIDE =  0, SIDEx =  0, SIDEy, SIDEz, SIDE_,
    UP   =  4, UPx   =  4, UPy,   UPz,   UP_,
    VIEW =  8, VIEWx =  8, VIEWy, VIEWz, VIEW_,
    POS  = 12, POSx  = 12, POSy,  POSz,  POS_,
  };
public:
  void move(M3Float dx, M3Float dy, M3Float dz)
    {set(POSx,get(POSx)+dx); set(POSy,get(POSy)+dy); set(POSz,get(POSz)+dz);}
  void move(M3Vector& step)
    {move(step.getX(),step.getY(),step.getZ());}
  void moveForward(M3Float step)
    {M3Vector dir(get(VIEWx),get(VIEWy),get(VIEWz)); move(dir.scale(step));}
  void moveStanding(M3Float step)
    {M3Vector dir(get(VIEWx),0,get(VIEWz)); move(dir.normalize().scale(step));}
  void roll(M3Float angle) {M3RotationZ rotZ(angle); rotateC(rotZ);}
  void yaw(M3Float angle) {M3RotationY rotY(angle); rotateC(rotY);}
  void pitch(M3Float angle) {M3RotationX rotX(angle); rotateC(rotX);}
  void rotateStanding(M3Float angle) {M3RotationY rotY(angle); rotate(rotY);}
  M3Float* getPositionArray() {return &matrix[POS];}
  M3Vector getPosition() {return M3Vector(get(POSx),get(POSy),get(POSz));}
  M3Float getPositionX() {return get(POSx);}
  M3Float getPositionY() {return get(POSy);}
  M3Float getPositionZ() {return get(POSz);}
  void setPosition(M3Vector& v)
    {set(POSx,v.getX()); set(POSy,v.getY()); set(POSz,v.getZ());}
  void setPosition(float x, float y, float z)
    {set(POSx,x); set(POSy,y); set(POSz,x);}
  void setPositionX(M3Float x) {set(POSx,x);}
  void setPositionY(M3Float y) {set(POSy,y);}
  void setPositionZ(M3Float z) {set(POSz,z);}
  M3Float* getViewDirectionArray() {return &matrix[VIEW];}
  M3Vector getViewDirection()
    {return M3Vector(get(VIEWx),get(VIEWy),get(VIEWz));}
  M3Float getViewDirectionX() {return get(VIEWx);}
  M3Float getViewDirectionY() {return get(VIEWy);}
  M3Float getViewDirectionZ() {return get(VIEWz);}
  void setViewDirection(M3Float x, M3Float y, M3Float z)
    {set(VIEWx,x); set(VIEWy,y); set(VIEWz,x);}
  M3Float* getUpDirectionArray() {return &matrix[UP];}
  M3Vector getUpDirection()
    {return M3Vector(get(UPx),get(UPy),get(UPz));}
  M3Float getUpDirectionX() {return get(UPx);}
  M3Float getUpDirectionY() {return get(UPy);}
  M3Float getUpDirectionZ() {return get(UPz);}
  void setUpDirection(M3Float x, M3Float y, M3Float z)
    {set(UPx,x); set(UPy,y); set(UPz,x);}
  M3Float* getSideDirectionArray() {return &matrix[SIDE];}
  M3Vector getSideDirection()
    {return M3Vector(get(SIDEx),get(SIDEy),get(SIDEz));}
  M3Float getSideDirectionX() {return get(SIDEx);}
  M3Float getSideDirectionY() {return get(SIDEy);}
  M3Float getSideDirectionZ() {return get(SIDEz);}
  void setSideDirection(M3Float x, M3Float y, M3Float z)
    {set(SIDEx,x); set(SIDEy,y); set(SIDEz,x);}
  void setTransform(
    M3Float x, M3Float y, M3Float z, M3Float theta, M3Float phi, M3Float rho
  );
  void setTransform(M3Vector& pos, M3Vector& orient) {
    setTransform(
      pos.getX(),pos.getY(),pos.getZ(),orient.getX(),orient.getY(),orient.getZ()
    );
  }
  M3Float* getTransform() {return getFloatArray();}
};

#endif //MATH3D_H
